/*
 * (C) Copyright IBM Corporation 2014, 2015.
*/
/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA

 Author: IBM Corporation and/or HCL Technologies Ltd.
 This module is part of the IBM (R) Rational (R) ClearCase (R)(*)
 Multi-version file system (MVFS).
 For support, please visit http://www.ibm.com/software/support

 * ClearCase is a trademark of IBM Corporation, registered in many
   jurisdictions, and is used under license.
*/
#ifndef _MVFS_LPC_K_H_
#define _MVFS_LPC_K_H_
/*
 * IBM Confidential
 * OCO Source Materials
 * Rational ClearCase
 * (C) Copyright IBM Corp. 2014, 2015.  All Rights Reserved.
 *
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 */
#include <mvfs_base.h>

/* Forward declare mfs_callinfo since mvfs_base.h includes this header and
 * mfs_callinfo is defined there.
 */
struct mfs_callinfo;

/* Defines an opaque handle to an LPC endpoint. Note that these handles are
 * reference counted.
 */
typedef KS_HANDLE(mvfs_lpc_h);

/* mvfs_lpc_hold
 *
 * References an LPC handle.
 *
 * Inputs:
 * lhp - A pointer to the lpc handle to reference.
 *
 * Return:
 * None
 */
#define mvfs_lpc_hold(lhp)  mvfs_rcr_hold(lhp)

/* mvfs_lpc_rele
 *
 * Releases an LPC handle.
 *
 * Inputs:
 * lhp - A pointer to the lpc handle to release.
 *
 * Returns:
 * None
 */
#define mvfs_lpc_rele(lhp) mvfs_rcr_rele(lhp)

/* mvfs_lpc_initialize
 *
 * Initializes the LPC subsystem. This function must be called before any of the
 * other LPC facilities can be used.
 *
 * Return:
 * 0 on success, errno otherwise.
 */
extern int mvfs_lpc_initialize(void);

/* mvfs_lpc_teardown
 *
 * Tears down the LPC subsystem. This function must be called before module
 * unload or memory leaks may result.
 *
 * Returns:
 * 0 on success, errno otherwise.
 */
extern int mvfs_lpc_teardown(void);

/* mvfs_lpc_create
 *
 * Inputs:
 * nm - An LPC endpoint identifier that will be used by userspace to locate and
 *      connect to the newly created endpoint.
 * cred - The cred of the user who owns this endpoint. Only the owner is allowed
 *        to connect to the endpoint (from userspace). A NULL cred indicates
 *        that anyone is allowed to connect to the endpoint.
 * lhp - A pointer to an mvfs_lpc_h that receives a handle to the newly created
 *       endpoint.
 *
 * XXXODFS: This function should be changed to take a pointer to a CALL_DATA_T
 * instead of a CRED_T to ensure that the LPC handle is always created using the
 * right credentials (the real user vs. possibly the system process) by getting
 * the credentials from the thread contained in the CALL_DATA_T (on Windows).
 */
extern int
mvfs_lpc_create(
    const char *nm,
    CRED_T *cred,
    mvfs_lpc_h *lhp
);

/* mvfs_lpc_call
 *
 * Makes an LPC (Local Procedure Call).
 *
 * Inputs:
 * lh - Specifies the LPC handle to make the call with.
 * trait - A pointer to the appropriate mfs_callinfo structure for the type of
 *         LPC being made.
 * op - Specifies the LPC operation number (synonymous with RPC operation
 *      number).
 * xid - Specifies the XID used to make the LPC with.
 * timeout - Specifies the limit for the duration of the LPC call (in tenths of
 *           a second, a value of zero indicates no timeout).
 * xdrargs - Specfies the XDR routine used to marshall the given argument
 *           structure with.
 * argsp - A pointer to an XDR'able argument structure for this LPC.
 * xdrres - Specifies the XDR routine used to unmarshal the LPC reply with.
 * resp - A pointer to a reply structure with receives the result of unXDR'ing
 *        the reply from the server.
 * vw - Specifies the view for this LPC, used to gather statistics.
 *
 * Return:
 * RPC_SUCCESS on success, one of the error codes defined in the clnt_stat enum
 * otherwise.
 */
extern enum clnt_stat
mvfs_lpc_call(
    mvfs_lpc_h lh,
    struct mfs_callinfo *trait,
    int op,
    XID_T xid,
    u_long timeout,
    xdrproc_t xdrargs,
    void *argsp,
    xdrproc_t xdrres,
    void *resp,
    VNODE_T *vw
);

#endif /* _MVFS_LPC_K_H_ */
/* $Id: 254c9626.310e11e5.8f9a.00:01:84:7a:f2:e4 $ */
