/*
 * (C) Copyright IBM Corporation 1990, 2016.
*/
/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA

 Author: IBM Corporation and/or HCL Technologies Ltd.
 This module is part of the IBM (R) Rational (R) ClearCase (R)(*)
 Multi-version file system (MVFS).
 For support, please visit http://www.ibm.com/software/support

 * ClearCase is a trademark of IBM Corporation, registered in many
   jurisdictions, and is used under license.
*/
/* mvfs_vfsops.c */
#include "mvfs_systm.h"
#include "mvfs.h"
#include "mvfs_dnc.h"
#include "mvfs_copy.h"

/* Static variable that includes all data used by vfs and statistics subsystems. */
mvfs_vfs_data_t mvfs_vfs_data_var;
mvfs_stats_data_t **mvfs_stats_data_ptr_percpu;

MVFS_MINMAP_T mvfs_minormap;	/* mvfs minor device map */

int mvfs_minmapsize; /* mvfs minor device map size */
MVFS_MAJOR_T *mvfs_majortbl; /* mvfs major number table */
#define VOBMOUNT_TABLESIZE \
    ((MVFS_MAJDYNMAX + MVFS_MAJFIXMAX) * MVFS_MINORMAX * sizeof(VFS_T *))

/* VFS mount generation number.  This is incremented on
   every unmount.  By using (fsid, gen) caches can be
   kept without any confusion when fsid's are reused because
   of unmounts. */

u_long mfs_mntgen = 0;

/* Our fs number and switch ptr */

MVFS_FSTYP_T	mfs_fstyp = 0;		/* MFS filesystem type */

V_OP_T  *mfs_vopp;		/* Ptr to mfs vnode ops switch */
VFSOPS_T *mfs_vfsopp;		/* Ptr to mfs vfs ops switch */
extern V_OP_T mvfs_vnodeops;

/* See the comments in mvfs_base.h regarding these variables and the rules
 * surrounding them.
 */
mvfs_state_t mvfs_state  = MDKI_ATOMIC_UINT32_INIT(MVFS_UNINITIALIZED);
MDKI_ATOMIC_UINT32_T mvfs_us_ops_in_progress = MDKI_ATOMIC_UINT32_INIT(0);

/* Defines how many times mvfs_unload attempts to unload before aborting. See
 * mvfs_unload for more details.
 */
#define MVFS_MAX_UNLOAD_ATTEMPTS 10

/* Turn on panics for Assertions by default */
tbs_boolean_t mvfs_panic_assert = FALSE;

/*
 * mvfs_max_cpus is the  maximum number of cpus that will ever be running
 * during this load of the mvfs.  It is set at mvfs load time.  It is used
 * by the statistics code to size the per-cpu statistics arrays.
 */

int mvfs_max_cpus;
tbs_boolean_t mvfs_cpu_beyond_limit;

#ifndef MVFS_SYSTEM_KMEM
extern struct mvfs_slab_list *mvfs_thread_slabs; 
extern struct mvfs_slab_list *mvfs_proc_slabs;
extern struct mvfs_slab_list *mvfs_cred_list_slabs; 
#endif

/* MFS init routine in SYS V.4 form.  */

/* Local function prototypes */
STATIC void
mvfs_misc_free(void);

STATIC void
mvfs_phase_1_teardown(void);

STATIC int
mvfs_vmount_subr(
    VFS_T *vfsp,
    VNODE_T *vwp,
    mfs_pn_char_t *path,
    int flags,
    caddr_t data,
    size_t datalen,
    CALL_DATA_T *acd,
    MVFS_CALLER_INFO *callinfo
);

STATIC int
mvfs_vunmount_subr(
    VFS_T *vfsp,
    CALL_DATA_T *cd
);

STATIC int
mvfs_find_devnum(
    int mmi_minor, 
    dev_t *majorp, 
    int *minorp
);

int
mvfsinit(
    VFSSW_T *vswp,
    int fstype
)
{
    int error = 0;
    mvfs_common_data_t *mcdp;
    mvfs_vfs_data_t *mvdp;
    mvfs_state_t state;

    /* Comments previously in this routine suggest that we can race
     * with unload and mount. We're going to account for this under the
     * revised sync/unload/mount/unmount synchronization scheme by
     * incrementing mvfs_us_ops_in_progress in the beginning of this
     * function and only changing mvfs_state to MVFS_PHASE_1_INITIALIZED at
     * the very end. This will prevent a racing unload from occuring until
     * the end because mvfs_us_ops_in_progress will be greater than zero. A
     * mount/sync race cannot occur because mvfs_state will remain
     * set to MVFS_UNINITIALIZED until all initialization is complete.
     */
    MDKI_ATOMIC_INCR_UINT32(&mvfs_us_ops_in_progress);

    if (MDKI_ATOMIC_READ_UINT32(&mvfs_state) != MVFS_UNINITIALIZED) {
        goto out;
    }

    /*
     * First, do any machine dependent init.
     * This is an optional macro in the mdep.h file for the port.
     */
    MVFS_MDEP_INIT();

    /*
     * adjust cache sizes might print, so init the lock here.
     */
    INITLOCK(&mvfs_printf_lock, "mvfspnrt");
    INITLOCK(&mvfs_printstr_lock, "mvfspstr");

    mcdp = MDKI_COMMON_GET_DATAP();
    BZERO(mcdp, sizeof(mvfs_common_data_t));

    mvfs_copy_tunable(mcdp);

    /* Initialize the global vfs data structure. */
    mvdp = MDKI_VFS_GET_DATAP(NULL);

    /* Initialize everything but the vob mount table for now. At this point, the
     * minormap (and hence MVFS_MINORMAX and other assorted tunables used to
     * initialize the vob mount table) hasn't been initialized. We'll set up the
     * rest of this structure on the first viewroot mount.
     */
    mvfs_vfs_data_init(MVFS_VDIF_ALL & ~MVFS_VDIF_VOB_MOUNTS, mvdp, "global");
    /* Display our version on initialization */
    mvfs_log(MFS_LOG_INFO, "%s", mvfs_version);

    /* Set up fs type and switch ptr */
    mfs_fstyp = (MVFS_FSTYP_T)fstype;

    MFS_VFSSW_ASSIGN(vswp, mfs_fstyp, mvfs_vfsops);
    mfs_vfsopp = &mvfs_vfsops;
    mfs_vopp = &mvfs_vnodeops;

#ifndef MVFS_SYSTEM_KMEM
    mfs_kmem_init(); /* Init kmem alloc (debug) before all else */
#endif

    MVFS_THREAD_SLAB_INIT(mvfs_thread_slabs,
                          sizeof(mvfs_thread_t),
                          TRUE,
                          "mvfs_thread_slabs");

    MVFS_PROC_SLAB_INIT(mvfs_proc_slabs,
                        sizeof(mvfs_proc_t),
                        TRUE,
                        "mvfs_proc_slabs");

#ifdef MVFS_VATTR_SLAB_ALLOC
    MVFS_VATTR_SLAB_INIT(mvfs_vattr_slabs,
                         sizeof(VATTR_T),
                         TRUE,
                         "mvfs_vattr_slabs");
#endif

    MVFS_CLR_CRED_SLAB_INIT(mvfs_cred_list_slabs,
                            sizeof(mvfs_clr_creds_t),
                            TRUE,
                            "mvfs_cl_slab");

retry_sw:
    switch (MDKI_ATOMIC_READ_UINT32(&mvfs_state)) {
      case MVFS_UNLOAD_IN_PROGRESS:
        /* There is a case where we could race with mvfs_unload. In this case,
         * do what unload would do to teardown everything we just set up and
         * leave mvfs_state as is since we're unloading.
         */
        mvfs_phase_1_teardown();
        error = EIO;
        break;
      case MVFS_UNINITIALIZED:
        if (!MDKI_ATOMIC_CAS_UINT32(&mvfs_state,
                                    MVFS_UNINITIALIZED,
                                    MVFS_PHASE_1_INITIALIZED))
        {
            /* It's possible that mvfs_unload could have been called in between
             * the read above and the previous CAS.
             */
            goto retry_sw;
        }
        break;
      default:
        /* This should never happen. The only questionable case is an init/init
         * race where this function has somehow already run to completion and
         * set mvfs_state to MVFS_PHASE_1_INITIALIZED. This, however, doesn't
         * seem possible (the old code didn't protect against init/init races
         * either).
         */
        MDKI_PANIC("mvfsinit: Incorrect mvfs_state on return.");
    }

out:
    if (error != 0) {
        mvfs_vfs_data_clean(mvdp);
    }

    MDKI_ATOMIC_DECR_UINT32(&mvfs_us_ops_in_progress);

    return(error);
}

int
mvfs_minormap_init(mvfs_cache_sizes_t *mma_sizes)
{

        int error;
        MDKI_SET_MINORMAX(MVFS_MINORMAX, MVFS_VIEW_SHIFT_BITS);
        MDKI_SET_VIEW_MASK_BITS(MVFS_VIEW_MASK_BITS, MVFS_VIEW_SHIFT_BITS);

	/* Init the minor map */
	MVFS_MAPINIT(&mvfs_minormap, error);
        if (error != 0)
            return error;

        /* Init the major device number table */
        error = MVFS_MAJORTBL_INIT(&mvfs_majortbl);
        if (error != 0) {
            MVFS_MAPFREE(&mvfs_minormap);
            return error;
        }
        
        return 0;
}

void
mvfs_minormap_free(void)
{

    MVFS_RELEASE_MAJORS(mvfs_majortbl);
    MVFS_MAJORTBL_FREE(mvfs_majortbl);
    MVFS_MAPFREE(&mvfs_minormap);
}

/* mvfs_vfs_get_datap
 * Returns a pointer to the corresponding vfs data structure for the given
 * vnode. This function always returns a valid pointer.
 *
 * Inputs:
 * vp - The vnode to retrieve the vfs data for.
 *
 * Returns:
 * A pointer to the vfs data corresponding to the given vnode or a pointer to
 * the global vfs data structure if vp is NULL.
 *
 */
mvfs_vfs_data_t *
mvfs_vfs_get_datap(VNODE_T *vp)
{
    /* In mvfs.h we define a number of C global data pointers for platforms that
    ** don't support zones (the MVFS "generic name" for Solaris zones, AIX
    ** wpar's, Linux and Windows containers, etc.).  For platforms that support
    ** zones, these same globals are put into an array of mvfs_zone_t entries
    ** indexed by zone (the global zone is index 0).  One of the data pointers
    ** is of mvfs_vfs_data_t and this routine is responsible for returning an
    ** appropriate mvfs_vfs_data_t pointer to its caller for the input vnode.
    ** It returns the global or zone mvfs_vfs_data_t pointer (as appropriate for
    ** the platform) in many cases, but for automatic views it returns the
    ** mvfs_vfs_data_t pointer that is stored in the vob_mnts field of the
    ** mfs_viewnode associated with the input vnode.  Since each mnode class has
    ** different ways to find a corresponding view (or a NULL vnode indicating
    ** the use of the non-automatic view data) we use a select statement and
    ** (one level of) recursion to get the right pointer.
    */
    if (vp == NULL) {
        goto return_nonauto;
    }

    if (!MFS_VPISMFS(vp)) {
        mvfs_log(MFS_LOG_WARN,
                 "mvfs_vfs_get_datap: Passed a non-MVFS vnode=%p"
                 " returning global vfs data pointer.\n",
                 vp);
        goto return_nonauto;
    }

    ASSERT(VTOM(vp) != NULL);

    switch (VTOM(vp)->mn_hdr.mclass) {
      case MFS_VOBCLAS:
        return mvfs_vfs_get_datap(MFS_VIEW(vp));
      case MFS_VOBRTCLAS:
        ASSERT(vp->v_vfsp != NULL);
        ASSERT(VFS_TO_MMI(vp->v_vfsp) != NULL);

        /* VOB roots have an embedded viewvp which is NULL for Dynamic Views and
         * filled in for Automatic Views.
         */
        return mvfs_vfs_get_datap(VFS_TO_MMI(vp->v_vfsp)->mmi_viewvp);
      case MFS_VOBTAGCLAS:
        ASSERT(VTOM(MFS_VIEW(vp))->mn_view.auto_view == 1);
        return mvfs_vfs_get_datap(MFS_VIEW(vp));
      case MFS_VIEWCLAS:
      case MFS_NTVWCLAS:
        if (MVFS_VRTYPE_VW(vp) == AUVWROOT) {
            return VTOM(vp)->mn_view.vob_mnts;
        }

        /* Fall through */
      case MFS_VIEWDIRCLAS:
        /*  Dynamic Views and viewroots always use the global vfs data */
        break;
      case MFS_SDEVCLAS:
      case MFS_LOOPCLAS:
      default:
        mvfs_log(MFS_LOG_DEBUG,
                 "mvfs_vfs_get_datap: Defaulting to global vfs data pointer "
                 "for vp = %p, mclass = %d\n",
                 vp,
                 VTOM(vp)->mn_hdr.mclass);
        break;
    }

  return_nonauto:
    return &mvfs_vfs_data_var;
}

/* mvfs_vfs_data_init
 * Initializes the given mvfs_vfs_data structure.  The mount_lock within
 * the structure is initialized using the given tag.  The name of the lock is of
 * the form "mount_lock [tag]".
 *
 * Inputs:
 * mvdif - A flag value specifying the members to initialize. 
 * mvdp - The mvfs_vfs_data structure to initialized.
 * mount_lock_tag - A string containing the desired mvfs_mountlock tag.
 *
 * Returns:
 * Non-zero if an error occurs, zero otherwise.
 */
int
mvfs_vfs_data_init(
    mvfs_vfs_data_init_flags_t mvdif,
    mvfs_vfs_data_t *mvdp,
    const char *mount_lock_tag
)
{
    char *mount_lock_name = NULL;
    size_t mount_lock_name_size = 0;
    int error = 0;

    if (mvdif & MVFS_VDIF_LOCK) {
        if (mount_lock_tag == NULL ||
            (STRCMP(mount_lock_tag, "") == 0))
        {
            INITLOCK(&(mvdp->mount_lock), "mount_lock");
        } else {
            mount_lock_name_size = STRLEN(mount_lock_tag) +
                                   STRLEN("mount_lock [") +
                                   2; /* Extra space for "]\0" */

            mount_lock_name = KMEM_ALLOC(mount_lock_name_size, KM_SLEEP);

            if (mount_lock_name == NULL) {
                /* I suppose we could use the default here but if this
                 * allocation failed then the vob mount table allocation below
                 * is most likely to fail so just error out here.
                 */
                error = ENOMEM;
                goto cleanup;
            }

            MVFS_SNPRINTF(mount_lock_name,
                          (int)mount_lock_name_size,
                          "%s [%s]",
                          "mount_lock",
                          mount_lock_tag);

            mount_lock_name[mount_lock_name_size - 1] = '\0';

            INITLOCK(&(mvdp->mount_lock), mount_lock_name);

            /* INITLOCK makes a copy of mount_lock_name, so free it here since
             * we no longer need it.
             */
            KMEM_FREE(mount_lock_name, mount_lock_name_size);
        }
    }

    if (mvdif & MVFS_VDIF_VOB_MOUNTS) {
        /*
         * Init the vob mount table (VOBMOUNT_TABLESIZE may not evaluate to a
         * constant, but rather to a computation using a tunable value.  So
         * this routine must be called after the tunables in VOBMOUNT_TABLESIZE
         * have been set.
         */
        mvdp->vob_mounts = KMEM_ALLOC(VOBMOUNT_TABLESIZE, KM_SLEEP);

        if (mvdp->vob_mounts == NULL) {
            error = ENOMEM;
            goto cleanup;
        }

        BZERO(mvdp->vob_mounts, VOBMOUNT_TABLESIZE);
    } else {
        mvdp->vob_mounts = NULL;
    }

    if (mvdif & MVFS_VDIF_VMD) {
        mvdp->mount_count = 0;
        mvdp->vob_mount_hwm = 0;
    }

  cleanup:

    if ((mvdif & MVFS_VDIF_LOCK) && error != 0) {
        FREELOCK(&(mvdp->mount_lock));
    }

    return error;
}

/* mvfs_vfs_data_clean
 * Cleans the given mvfs_vfs_data structure, freeing the mount_lock and vob
 * mount table (if necessary).
 *
 * Inputs:
 * mvdp - The mvfs_vfs_data structure to clean.
 *
 * Returns:
 * void
 */
void
mvfs_vfs_data_clean(mvfs_vfs_data_t *mvdp)
{
    ASSERT(mvdp->mount_count == 0);

    if (mvdp->mount_count > 0) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_vfs_data_clean: Cleaning mvfs_vfs_data "
                 "structure (%"KS_FMT_PTR_T") where mount_count = %d. "
                 "Continuing...\n",
                 mvdp,
                 mvdp->mount_count);
    }

    FREELOCK(&(mvdp->mount_lock));

    if (mvdp->vob_mounts != NULL) {
        KMEM_FREE(mvdp->vob_mounts,
                  ((MVFS_MAJDYNMAX + MVFS_MAJFIXMAX) *
                   MVFS_MINORMAX *
                   sizeof(VFS_T *)));

        mvdp->vob_mounts = NULL;
    }
}

/* 
 * Clean up various structures no longer needed once we unmount all VOBs and
 * viewroot. 
 */
STATIC void
mvfs_misc_free(void)
{
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();

    /* 
     * Time to unload our data structures
     */

    MVFS_FREE_CREDLIST();
    mvfs_clnt_destroy();
    mvfs_acl_free();
    mvfs_dncfree();

    mvfs_viewfree();
    /*
     *  We must unload the audit after proc stuff: proc stuff might
     *  release things via audit, and that requires manipulating the
     *  audit locks that would be freed by mvfs_auditfree().
     */
    mvfs_procdata_free();
    mvfs_auditfree();
    mvfs_mndata_free();
    mvfs_util_free();
    mvfs_minormap_free();
    MDKI_STATS_FREE_DATA();
    MVFS_MDEP_MISC_FREE();
    mcdp->mvfs_largeinit = mcdp->mvfs_init_sizes.size[MVFS_SETCACHE_LARGEINIT];

    return;
}

/* Unload the file system.  If vswp is NULL, this is a non-VFSSW platform. */
int
mvfs_unload(
    VFSSW_T *vswp,
    int fsnum
)
{
    int num = 0;
    ks_uint32_t lstate;

retry_sw:
    if (num == MVFS_MAX_UNLOAD_ATTEMPTS) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_unload: Max unload attempts reached while trying to "
                 "perform transition to MVFS_UNLOAD_IN_PROGRESS.\n");
        return(EBUSY);
    }

    lstate = MDKI_ATOMIC_READ_UINT32(&mvfs_state);

    switch (lstate) {
      case MVFS_SYNC_IN_PROGRESS:
        MDKI_USECDELAY(50000); /* 0.05 seconds */
        ++num;
        goto retry_sw;
      case MVFS_FULLY_INITIALIZED:
      case MVFS_UNLOAD_IN_PROGRESS:
        mvfs_log(MFS_LOG_INFO,
                 "MVFS: mvfs_unload (%s): incorrect state to unload (%d)\n",
                 MVFS_CODE_LOC,
                 lstate);
        return(EBUSY);
      default: /* MVFS_UNINITIALIZED or MVFS_PHASE_1_INITIALIZED */
        if (!MDKI_ATOMIC_CAS_UINT32(&mvfs_state,
                                    lstate,
                                    MVFS_UNLOAD_IN_PROGRESS))
        {
            MDKI_USECDELAY(50000);
            ++num;
            goto retry_sw;
        }
    }

    while (MDKI_ATOMIC_READ_UINT32(&mvfs_us_ops_in_progress) > 0) {
        if (num++ == MVFS_MAX_UNLOAD_ATTEMPTS) {
            mvfs_log(MFS_LOG_INFO,
                     "mvfs_unload: Max unload attempts reached while waiting "
                     "on mvfs_us_ops_in_progress.\n");
        }

        MDKI_USECDELAY(50000);
    }

    /* Check for any active mnodes.  (Note: if there are no
     * mnodes, then there can't be any mounts since a mount needs
     * a root mnode at least!
     */
    if ((num = mfs_mncount()) > 0) {
        mvfs_log(MFS_LOG_INFO,
                 "MVFS: mvfs_unload: %d mnodes still open\n", num);
        return(EBUSY); /* Still busy */
    }

    /* Clear out our VFS ops ptr */
    if (vswp != NULL) {
        MFS_VFSSW_CLEAR(vswp, fsnum);
    }

    if (lstate == MVFS_PHASE_1_INITIALIZED) {

        MDKI_USECDELAY(1000000*2);
        mvfs_phase_1_teardown();
    }

cleanup:
    if (!MDKI_ATOMIC_CAS_UINT32(&mvfs_state,
                                MVFS_UNLOAD_IN_PROGRESS,
                                MVFS_UNINITIALIZED))
    {
        MDKI_PANIC("mvfs_unload: mvfs_state not MVFS_UNLOAD_IN_PROGRESS "
                   "after unload procedures completed\n");
    }

    return(0);
}

void
mvfs_phase_1_teardown(void)
{
    MVFS_KMEM_UNLOAD();

    /* Free the memory allocated for proc/thread slab management.
     * Separate from proc/thread hash tables because those come and go
     * with viewroot mounts.
     */
#ifndef MVFS_SYSTEM_KMEM
    MDB_XLOG((MDB_MEMOP,
              "proc/thread slab memory being freed: %"KS_FMT_PTR_T
              " %"KS_FMT_PTR_T"\n",
              mvfs_proc_slabs,
              mvfs_thread_slabs));
#endif /* XXX could put down 3 lines, but macros handle ... */
    MVFS_PROC_SLAB_DESTROY(mvfs_proc_slabs);
    MVFS_PROC_SLAB_DESTROY(mvfs_thread_slabs);
    MVFS_CLR_CRED_SLAB_DESTROY(mvfs_cred_list_slabs);

#ifdef MVFS_VATTR_SLAB_ALLOC
    MVFS_VATTR_SLAB_DESTROY(mvfs_vattr_slabs);
#endif

/* The vfs data for zones (including the global zone) are managed as part of
 * the zone creation/destruction process.
 */
    mvfs_vfs_data_clean(MDKI_VFS_GET_DATAP(NULL));

    FREE_INIT_LOCKS();
    MVFS_MDEP_UNLOAD();
}

/*
 * This function auto adjusts mvfs_largeinit based on system memory size.
 * Called if user has not adjusted the default value.  Generally, we try
 * to keep the old values of mvfs_largeinit for smaller desktop systems
 * while trying to set reasonable defaults for large Build Servers with
 * multiple-GB of memory.
 *  Algorithm:
 *  mvfs_largeinit is set with a step function with steps of width 
 *  MVFS_LARGEINITBLK.  However, the beginning is special-cased for
 *  compatibility with pre-5.0 settings of 0 and 1.  Then there is a
 *  long step with mvfs_largeinit of 2, to bridge the gap with large systems.
 *  After this, when memory size increases a step -- MVFS_LARGEINITBLK, 
 *  mvfs_largeinit will increase by 1, until it reaches MVFS_LARGEINITMAX.
 *  We never automatically size mvfs_largeinit above MVFS_LARGEINITMAX, 
 *  however, on UNIX, the user is free to set mvfs_largeinit to a larger number
 *  if they wish.
 */
void
mvfs_compute_largeinit(u_long kbytes_available)
{
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();
    /* keep the old settings for memory < (MVFS_LARGEINITBLK/2) */
    if (kbytes_available < MVFS_MINIMUM_LARGEMEM) {
        mcdp->mvfs_computed_largeinit = 0;
    }
    else if (kbytes_available < MVFS_LARGEINITBLK/2) {
        mcdp->mvfs_computed_largeinit = 1;
    }
    else if (kbytes_available < MVFS_LARGEINITBLK*2) {
        mcdp->mvfs_computed_largeinit = 2;
    }
    else {
        mcdp->mvfs_computed_largeinit = kbytes_available / MVFS_LARGEINITBLK;
        /* Restrict largeinit value to MVFS_LARGEINITMAX */
        if (mcdp->mvfs_computed_largeinit > MVFS_LARGEINITMAX) {
            mcdp->mvfs_computed_largeinit = MVFS_LARGEINITMAX;
        }
    }

}

void
mvfs_physmem_adj_caches(mvfs_cache_sizes_t *mma_sizes)
{
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();
    u_long kbytes_available; /* Physical memory size in Kilo-Bytes */

    /* Auto scale what mvfs_largeinit would be if not tuned by mma_sizes */
    /* Convert number of memory pages to kbytes. */
    kbytes_available = (MDKI_PAGESIZE() / 1024) * MVFS_PHYSMEM_PAGECNT;

    mvfs_compute_largeinit(kbytes_available);

    if (MVFS_SIZE_VALID(mma_sizes, LARGEINIT))
        MVFS_SIZE_LOAD(mcdp->mvfs_largeinit, mma_sizes, LARGEINIT);
    else {
        mcdp->mvfs_largeinit = mcdp->mvfs_computed_largeinit;
        mvfs_log(MFS_LOG_INFO,
                 "MVFS: Setting mvfs_largeinit to %d based on physical memory"
                 " size of %luKB\n", mcdp->mvfs_largeinit, kbytes_available);
    }
}

/*
 * Complete any deferred initialization, using mount args to set
 * sizes, etc.  (It's put here in the file to avoid having to add
 * prototypes for some functions it calls and are defined above.)
 */
int
mvfs_misc_init(
    mvfs_cache_sizes_t *mma_sizes
)
{
    int error, cpuid;
    struct cpu *cp, *cur_cpu;

    mvfs_max_cpus = MVFS_GET_MAXCPU;
    mvfs_cpu_beyond_limit = FALSE;

    MDKI_STATS_ALLOC_DATA();

    for (cpuid = 0; cpuid < mvfs_max_cpus; cpuid++) {
        MDKI_STATS_GET_DATAP(cpuid) = NULL;
    }

    /*
     * For the statistics structures, memory is allocated dynamically when
     * a process runs on a CPU for the first time.  This is handled in the
     * statistics manipulation routines, e.g. BUMPSTAT, DECSTAT, etc .
     */

    /*
     * adjust cache memory sizes
     */
    MVFS_ADJUST_CACHESIZES(mma_sizes);

    error = MVFS_MDEP_MISC_INIT(mma_sizes);
    if (error != 0)
        return error;
    error = mvfs_minormap_init(mma_sizes);
    if (error != 0)
        goto mdeperr;
    error = mvfs_util_init(mma_sizes); /* initialize utility things */
    if (error != 0)
        goto maperr;
    error = mvfs_mninit(mma_sizes); /* Init mnodes */
    if (error != 0) {
        goto utilerr;
    }
    error = mvfs_procinit(mma_sizes); /* Init per-process info tables */
    if (error != 0) {
        goto mnerr;
    }
    error = mvfs_auditinit(mma_sizes); /* Init audit structs */
    if (error != 0) {
        /*
         * should have no auditing set up yet, so it's safe to
         * unload without having audit initialized yet
         */
        mvfs_procdata_free();
        goto mnerr;
    }
    error = mvfs_viewinit(mma_sizes); /* Init view structs */
    if (error != 0) {
        goto procerr;
    }
    error = mvfs_dncinit(mma_sizes);   /* Init dir name cache */
    if (error != 0) {
        goto viewerr;
    }
    error = mvfs_clnt_init(mma_sizes);	/* Init rpc client cache */
    if (error != 0) {
        goto dncerr;
    }
    error = mvfs_acl_init(); /* Init the EACL cache. */
    if (error != 0) {
        goto clnterr;
    }
    error = MVFS_INIT_CREDLIST(mma_sizes); /* Init cleartext cred cache */
    if (error != 0) {
        /* Unwind in reverse order */
        /* don't need to undo MVFS_INIT_CREDLIST, it failed so we came here */
        mvfs_acl_free();
      clnterr:
        mvfs_clnt_destroy();
      dncerr:
        mvfs_dncfree();
      viewerr:
        mvfs_viewfree();
      procerr:
        /*
         *  We must unload the audit after proc stuff: proc stuff might
         *  release things via audit, and that requires manipulating the
         *  audit locks that would be freed by mvfs_auditfree().
         */
        mvfs_procdata_free();
        mvfs_auditfree();
      mnerr:
        mvfs_mndata_free();
      utilerr:
        mvfs_util_free();
      maperr:
        mvfs_minormap_free();
      mdeperr:
        MVFS_MDEP_MISC_FREE();
    }
    return error;
}

/*
 * Allocate memory for the statistics on the specified CPU and initialize
 * the stats structure.  Called from the statistics manipulation routines,
 * e.g. BUMPSTAT, DECSTAT, etc.
 */
mvfs_stats_data_t *
mvfs_stats_data_per_cpu_init(void)
{
    /* Initialize the statistics structure for the current cpu */
    mvfs_stats_data_t *sdp =
        (mvfs_stats_data_t *)KMEM_ALLOC(sizeof(mvfs_stats_data_t), KM_SLEEP);

    if (sdp == NULL) {
        mvfs_log(MFS_LOG_ERR,
                 "Failed to allocate memory for statistics: ENOMEM\n");
    } else {
        MVFS_PERCPU_STAT_ZERO(sdp);
        sdp->zero_me = FALSE;
    }

    return(sdp);
}

VFS_T *
mvfs_findvfs_nm(
    mvfs_vfs_data_t *mvdp,
    tbs_boolean_t case_insensitive,
    mfs_pn_char_t *nm
)
{
    int i;
    VFS_T *vfsp;
    int myzoneid = MDKI_GETZONEID();

    ASSERT(ISLOCKED(&(mvdp->mount_lock)));

    for (i = 0; i <= mvdp->vob_mount_hwm; i++) {
        vfsp = mvdp->vob_mounts[i];

        if (vfsp == NULL) {
            continue;
        }

        ASSERT(myzoneid == VFS_TO_MMI(vfsp)->mmi_zoneid);

        if (PN_STRCMP(case_insensitive,
                      VFS_TO_MMI(vfsp)->mmi_vobtag,
                      nm) == 0)
        {
            return(vfsp);
        }
    }

    return(NULL); /* Not found */
}

VFS_T *
mvfs_findvfs_mntpath(
    mvfs_vfs_data_t *mvdp,
    mfs_pn_char_t *path
)
{
    int i;
    VFS_T *vfsp;
    int myzoneid = MDKI_GETZONEID();

    ASSERT(ISLOCKED(&(mvdp->mount_lock)));

    for (i = 0; i <= mvdp->vob_mount_hwm; i++) {
        vfsp = mvdp->vob_mounts[i];

        if (vfsp == NULL) {
            continue;
        }

        /* FIXME: findvfs_mntpath should support case-insensitive someday */
        if (myzoneid == VFS_TO_MMI(vfsp)->mmi_zoneid &&
            PN_STRCMP(FALSE, VFS_TO_MMI(vfsp)->mmi_mntpath, path) == 0)
        {
            return(vfsp);
        }
    }

    return(NULL); /* Not found */
}

VFS_T *
mvfs_findvfs_oid(
    mvfs_vfs_data_t *mvdp,
    tbs_oid_t *oidp,
    int *indx,
    tbs_boolean_t *unique
)
{
    int i = 0;
    VFS_T *vfsp;
    VFS_T *saved_vfsp = NULL;
    int myzoneid = MDKI_GETZONEID();

    *unique = TRUE;

    ASSERT(ISLOCKED(&(mvdp->mount_lock)));

    if (indx) {
        i = *indx;
    }

    for (; i <= mvdp->vob_mount_hwm; i++) {
        vfsp = mvdp->vob_mounts[i];

        if (vfsp == NULL) {
            continue;
        }
        if (MFS_OIDEQ(VFS_TO_MMI(vfsp)->mmi_voboid, *oidp)) {
            if (indx) {
                *indx = i + 1;
                indx = NULL;
            }

            if (saved_vfsp) {
                *unique = FALSE;
            } else {
                saved_vfsp = vfsp;
                ASSERT(myzoneid == VFS_TO_MMI(vfsp)->mmi_zoneid);
            }
        }
    }

    return(saved_vfsp);
}

VFS_T *
mvfs_findvfs_uuid(
    mvfs_vfs_data_t *mvdp,
    tbs_uuid_t *uuidp
)
{
    VFS_T *vfsp;
    int i;
    int myzoneid = MDKI_GETZONEID();

    ASSERT(ISLOCKED(&(mvdp->mount_lock)));

    for (i = 0; i <= mvdp->vob_mount_hwm; i++) {
        vfsp = mvdp->vob_mounts[i];

        if (vfsp == NULL) {
            continue;
        }

        if (MFS_UUIDEQ(VFS_TO_MMI(vfsp)->mmi_vobuuid, *uuidp)) {
            ASSERT(myzoneid == VFS_TO_MMI(vfsp)->mmi_zoneid);
            return(vfsp);
        }
    }

    return(NULL); /* Not found */
}

VFS_T *
mvfs_findvfs_cookie(
    mvfs_vfs_data_t *mvdp,
    u_long *cookiep
)
{
    VFS_T *vfsp;
    u_long i;
    int myzoneid = MDKI_GETZONEID();

    ASSERT(ISLOCKED(&(mvdp->mount_lock)));

    for (i = *cookiep; i <= mvdp->vob_mount_hwm; i++) {
        vfsp = mvdp->vob_mounts[i];

        if (vfsp == NULL) {
            continue;
        }

        *cookiep = i + 1; /* Leave cookie at next entry */
        ASSERT(myzoneid == VFS_TO_MMI(vfsp)->mmi_zoneid);
        return(vfsp);
    }

    return(NULL); /* No more vob mounts found */
}

/*
 * Main mount routines
 */
int
mvfs_vmount(
    VFS_T *vfsp,    /* Our vfs ptr (mount tab) */
    VNODE_T *vwp,   /* if NULL, viewroot/global VOB, else view vp for au VOB. */
    char *path,     /* Pathname of special file */
    int flags,      /* flags to the mount */
    caddr_t data,   /* data args */
    size_t datalen, /* data args length */
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo /* Caller info (irp info)    */
)
{
    mvfs_vfs_data_t *mvdp = MDKI_VFS_GET_DATAP(vwp);
    int error;

    /* Hold the appropriate mount_lock ("global" for viewroots or global VOB
    ** mounts, per-auview for automatic view VOB mounts).
    */
    MVFS_LOCK(&(mvdp->mount_lock));

    error = mvfs_vmount_subr(vfsp,
                             vwp,
                             path,
                             flags,
                             data,
                             datalen,
                             cd,
                             callinfo);

    MVFS_UNLOCK(&(mvdp->mount_lock));

    return error;
}

/*
 * MFS register/unregister calls.  Actually, this allocates
 * a bogus vfsp, and then calls mount with the normal
 * mount arguments.  It can be used on either Unix or NT
 * to support vobs under Nt-style view-tags without any
 * real mounting required.
 */

int
mvfs_register_mount(
    mvfs_viewroot_type_t vrtype,
    VNODE_T *vwp,   /* if NULL, viewroot/global VOB, else view vp for au VOB. */
    caddr_t mnt_data,
    size_t mnt_datalen,
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo
)
{
    VFS_T *vfsp = NULL;
    int error = 0;
    int error2;                    /* Error for logging, not returning. */
    tbs_boolean_t mounted = FALSE; /* For cleanup. */

    /* Make sure our arguments are consistent (as best we can)...Either this is
    ** a global mount of a viewroot or VOB, or it is a per-auview VOB mount.  We
    ** can't check that it is a VOB mount (of either type, rather than a
    ** viewroot mount) because that is "encoded" in the mount data, which is
    ** parsed by the mvfs_vmount call below.  So just check what we can: if it
    ** looks like an auview VOB mount (vwp != NULL), then its an error if the
    ** vrtype is not AUVWROOT or the vwp is not for an automatic view.
    */
    if (vwp != NULL && ((vrtype != AUVWROOT) || !MVFS_VIEW_IS_AUTO_VIEW(vwp))) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_register_mount: Invalid vrtype=%s and vwp=%p"
                 " MFS_VPISMFS=%d"
                 " MFS_ISNTVIEW=%d (mclass=%d)"
                 " MVFS_VRTYPE_VW=%s\n",
                 MVFS_VWROOT_TO_STR(vrtype), vwp,
                 MFS_VPISMFS(vwp),
                 MFS_VPISMFS(vwp) ? MFS_ISNTVIEW(VTOM(vwp)) : -1,
                 MFS_VPISMFS(vwp) ? VTOM(vwp)->mn_hdr.mclass : -1,
                 MFS_VPISMFS(vwp) && MFS_ISNTVIEW(VTOM(vwp)) ?
                     MVFS_VWROOT_TO_STR(MVFS_VRTYPE_VW(vwp)) :
                     MVFS_VWROOT_TO_STR(NOVWROOT));
        error = EINVAL;
        goto done;
    }
    /* Allocate a vfs struct for the mount */
    if ((vfsp = (VFS_T *)KMEM_ALLOC(sizeof(VFS_T), KM_SLEEP)) == NULL) {
        error = ENOMEM;
        goto done;
    }
    /*
     * Initialize the VFS structure.
     * We use the VFS op ptr for MVFS VFS ops set up in mvfsinit()
     * We do not assign a data ptr yet, the mount subroutine does this later.
     */
    MVFS_INITVFS(vfsp, mfs_vfsopp, NULL);

    if ((error = mvfs_vmount(vfsp, vwp, NULL, 0, mnt_data, mnt_datalen, cd,
                             callinfo)) != 0)
    {
        goto done;
    }
    mounted = TRUE;             /* Cleanup on a subsequent error. */

    /* If this is an automatic view VOB mount we need to keep track of the
    ** vobtag components in a per-view data structure so we can support the
    ** components as "virtual directories".
    */
    if (vwp != NULL) {
        /* Now we can check that it is a VOB mount and the VFS is correct. */
        if (!MFS_ISVOBMNT(VFS_TO_MMI(vfsp)) ||
            (vwp != VFS_TO_MMI(vfsp)->mmi_viewvp))
        {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_register_mount: Not automatic view VOB mount"
                     " vwp=%p vfsp=%p mmi_viewvp=%p mmi_mnttype=%d\n",
                     vwp,
                     vfsp,
                     VFS_TO_MMI(vfsp)->mmi_viewvp,
                     VFS_TO_MMI(vfsp)->mmi_mnttype);
            error = EINVAL;
            goto done;
        }
        /* Fill in the per-auview "virtual directory" of vobtag components. */
        if ((error = mvfs_auvw_mount_vobtag(vfsp, cd)) != 0) {
            goto done;
        }
    }

  done:
    if (error != 0) {
        /* Cleanup the mount, if necessary. */
        if (mounted && (error2 = mvfs_vunmount(vfsp, cd)) != 0) {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_register_mount: Calling mvfs_vunmount(vfsp=%p)"
                     " after err=%d%s got err2=%d%s...Continuing\n",
                     vfsp,
                     error, mfs_strerr(error),
                     error2, mfs_strerr(error2));
            /* Things are still wrong, but we tried, so continue. */
        }
        /* Cleanup what we allocated. */
        if (vfsp != NULL) {
            KMEM_FREE(vfsp, sizeof(*vfsp));
        }
    }
    return error;
}

/****************************************************************************
 * mvfs_unregister_mount
 * Unmount ioctl (was mvfs_nt2vfs_umount).  Only for NT-style views.
 * IN   vrtype          viewroot type that received the unmount ioctl.
 * IN   vp              view vnode for the view to do the unmount in.
 * IN   unmount_ptr     ptr to unmount information
 * IN   cred            ptr to the cred
 * RESULT:              error code
 */
int
mvfs_unregister_mount(
    mvfs_viewroot_type_t vrtype,
    VNODE_T *vwp,
    struct mvfs_unmount_info *unmount_ptr,
    CALL_DATA_T *cd
)
{
    VFS_T *vfsp;
    int error;
    mvfs_vfs_data_t *mvdp;

    ASSERT(vwp == NULL || (MFS_VPISMFS(vwp) && MFS_ISNTVIEW(VTOM(vwp))));
    mvdp = MDKI_VFS_GET_DATAP(vwp);

    /* Find VFSP from replica uuid */
    MVFS_LOCK(&(mvdp->mount_lock)); /* Lock for findvfs core routine */

    if ((vfsp = mvfs_findvfs_uuid(mvdp, &unmount_ptr->replica_uuid)) == NULL) {
        error = EINVAL;
    } else {
        /* Do the unmount operation */
        if ((error = mvfs_unregister_mount_subr(vfsp, cd)) == EBUSY) {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_unregister_mount: unregister mount failed EBUSY"
                     " - retrying\n");
            /* Do one retry */
            error = mvfs_unregister_mount_subr(vfsp, cd);
        }
        if (error) {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_unregister_mount: failed err=%d%s\n",
                     error, mfs_strerr(error));
        }
    }
    MVFS_UNLOCK(&(mvdp->mount_lock));

    return error;
}

/* This routine expects to be called with the mount_lock corresponding
 * to the given vfs to be locked.
 */
int
mvfs_unregister_mount_subr(
    VFS_T *vfsp,
    CALL_DATA_T *cd
)
{
    int error;
    int error2;                 /* Error for logging, not returning. */
    tbs_boolean_t vobtag_dir_clean = FALSE;

    /* For automatic views, if the mvfs_vunmount_subr below succeeds, the vfs
    ** will have been cleaned up (mmi_rootvp = NULL, etc.), so clean up our
    ** vobtag "virtual directory" data first.  If the vunmount fails, we'll have
    ** to put our stuff back because the vfs will still be there.
    */
    if (MVFS_VRTYPE_VFS(vfsp) == AUVWROOT) {
        if ((error = mvfs_auvw_unmount_vobtag(vfsp, cd)) != 0) {
            goto done;
        }
        vobtag_dir_clean = TRUE; /* For cleanup below. */
    }

    if ((error = mvfs_vunmount_subr(vfsp, cd)) != 0) {
        goto done;
    }
    /* Now that we are unmounted, free up the VFS struct and
     * clean up any state set up by MVFS_INITVFS, e.g. lock cleanup
     */
    MVFS_FINISHVFS(vfsp);
    KMEM_FREE(vfsp, sizeof(*vfsp));
    ASSERT(error == 0);         /* Verify we won't use vfsp anymore. */

  done:
    if ((error != 0) && vobtag_dir_clean) {
        /* Put back the vobtag information. */
        if ((error2 = mvfs_auvw_mount_vobtag(vfsp, cd)) != 0) {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_unregister_mount_subr:"
                     " Calling mvfs_auvw_mount_vobtag(vfsp=%p)"
                     " after err=%d%s got err2=%d%s...Continuing\n",
                     vfsp,
                     error, mfs_strerr(error),
                     error2, mfs_strerr(error2));
            /* Things are still wrong, but we tried, so continue. */
        }
    }
    return error;
}

int
mvfs_unregister_all_vobs(
    mvfs_viewroot_type_t vrtype,
    VNODE_T *vp,
    CALL_DATA_T *cd
)
{
    int i;
    VFS_T *vfsp;
    int error;

    /*
     * This var "some_error" is set non-zero if some unmount has
     * an error.  It is the "last" unmount error encountered.
     */
    int some_error = 0;
    mvfs_vfs_data_t *mvdp = MDKI_VFS_GET_DATAP(vp);

    /* XXXODFS: Maybe we want to unmount all VOBs in all views for a given user
     * if vrtype is AUVWROOT and vp is NULL?
     */
    if (vrtype == AUVWROOT && vp == NULL) {
        mvfs_log(MFS_LOG_DEBUG,
                 "mvfs_unregister_all_vobs: Called for automatic viewroot "
                 "with no view context.");
        return EINVAL;
    }

    MVFS_LOCK(&(mvdp->mount_lock));

    for (i = 0; i <= mvdp->vob_mount_hwm; i++) {
        vfsp = mvdp->vob_mounts[i];
        if (vfsp == NULL) {
            continue;
        }

        error = mvfs_unregister_mount_subr(vfsp, cd);

        if (error) {
            some_error = error;
            mvfs_log(MFS_LOG_WARN,
                     "trouble unregistering mount %d%s\n",
                     i,
                     mfs_strerr(error));
        }
        /* Continue after errors */
    }

    MVFS_UNLOCK(&(mvdp->mount_lock));

    return(some_error);
}

STATIC int MVFS_NOINLINE
mvfs_viewroot_common_init(
    VFS_T *vfsp,
    mvfs_viewroot_type_t vrtype,
    struct mfs_mntargs *mmap,
    struct mfs_mntinfo *mmip,
    CALL_DATA_T *cd
)
{
    int error;
    dev_t major_num;
    int minor_num;
    VNODE_T *rtvp;              /* viewroot vp. */
    VNODE_T *svp;               /* .specdev vp. */
    int num;                    /* mfs_ramdir_add slot, ignored. */
    mvfs_viewroot_data_t *vrdp = MDKI_VIEWROOT_GET_DATAP(vrtype);

    /* Only allow one mount of this viewroot. */
    if (vrdp->mfs_viewroot_vfsp != NULL) {
        mvfs_log(MFS_LOG_ERR,
                 "The %s viewroot is already mounted.\n",
                 MVFS_VWROOT_TO_STR(vrtype));
        error = EEXIST;
        goto errout;
    }
    /* Set up MMI info before any makenodes. */
    mmip->mmi_minor = mvfs_vfsgetnum(&mvfs_minormap, MVFS_MINMAPSIZE);
    mmip->mmi_nvminor = -1;  /* Not used for viewroot mountpoint */
    mmip->mmi_gen = mfs_mntgen;

    vfsp->vfs_bsize = DEV_BSIZE; /* XXXODFS Do we need this? */

    /* For Linux, the device number is assigned before we are called */
    MVFS_SET_FSTYPE(vfsp);
    MVFS_ATTACH_VFS_DATA(vfsp, mmip);
    if ((error = mvfs_find_devnum(mmip->mmi_minor, &major_num, &minor_num))
        != 0)
    {
        goto errout;
    }
    MAKEFSID(VFS_FSID(vfsp), major_num, minor_num);
    MAKEVFSDEV(vfsp, major_num, minor_num, major_num, -1);

    /* Set up the mount point history mode suffix if passed in. */
    if ((mmap->mma_hmsuffix.s != NULL) && (mmap->mma_hmsuffix.l) > 0) {
        if ((error = mfs_copyin_strbuf(mmap->mma_hmsuffix,
                                       &(mmip->mmi_hmsuffix))) != 0)
        {
            goto errout;
        }
        mmip->mmi_hmsuffixlen = STRLEN(mmip->mmi_hmsuffix);
    } else {
        mmip->mmi_hmsuffix = NULL;
        mmip->mmi_hmsuffixlen = 0;
    }
    /*
     * Set up the mount point HM version name if there is a hm suffix
     * configured.  Normally .@@ will "warp" you to the element level, so you
     * need to say .@@/main/LATEST/file.c/main/LATEST.  The hm version name can
     * be used to warp into hm to the version of the current directory so that
     * you can say: ^@@/file.c/main/LATEST.
     */
    if (mmip->mmi_hmsuffix) {
        /* Allocate suffixlen + 2 for extra char and trailing null */
        mmip->mmi_hmvers_nm = KMEM_ALLOC(mmip->mmi_hmsuffixlen + 2, KM_SLEEP);
        if (mmip->mmi_hmvers_nm != NULL) {
            mmip->mmi_hmvers_nm[0] = MVFS_VX_VERS_CHAR; /* Canned ... */
            STRCPY(&mmip->mmi_hmvers_nm[1], mmip->mmi_hmsuffix);
        } else {
            error = ENOMEM;
            goto errout;
        }
    }
    /* If there are multiple viewroots, check to see that the hmsuffix is the
    ** same.  Log an error if it isn't, but keep going...maybe the user really
    ** meant to do it :)
    */
    if ((MDKI_VIEWROOT_GET_DATAP(DYVWROOT)->mfs_viewroot_vfsp != NULL) &&
        (MDKI_VIEWROOT_GET_DATAP(AUVWROOT)->mfs_viewroot_vfsp != NULL))
    {
        struct mfs_mntinfo *dymmip =
            VFS_TO_MMI(MDKI_VIEWROOT_GET_DATAP(DYVWROOT)->mfs_viewroot_vfsp);
        struct mfs_mntinfo *odmmip =
            VFS_TO_MMI(MDKI_VIEWROOT_GET_DATAP(AUVWROOT)->mfs_viewroot_vfsp);

        if ((odmmip->mmi_hmsuffix != NULL) &&
            (dymmip->mmi_hmsuffix != NULL) &&
            STRCMP(odmmip->mmi_hmsuffix, dymmip->mmi_hmsuffix) != 0)
        {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_viewroot_common_init: automatic view hmsuffix: %s"
                     " and dynamic view hmsuffix: %s"
                     " do not match...continuing.\n",
                     odmmip->mmi_hmsuffix, dymmip->mmi_hmsuffix);
        }
    }
    /* Set up the kind of view-tags to create under this mount.
    ** FIXME: for now, based on the PORT.  But...should be a mount-argument...
    */
    if (vrtype == DYVWROOT) {
        mmip->mmi_default_vwtag_kind = MFS_VIEWCLAS;
    } else {
        ASSERT(vrtype == AUVWROOT);
        mmip->mmi_default_vwtag_kind = MFS_NTVWCLAS;
    }

    /* Make the viewroot node (this allocates the ramdir, too). */
    if ((error = mfs_makespecnode(MFS_VIEWDIRCLAS, (VNODE_T *)NULL,
                                  (void *)NULL, vfsp, &rtvp, cd)) != 0)
    {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_viewroot_common_init: making %s viewroot err=%d%s\n",
                 MVFS_VWROOT_TO_STR(vrtype), error, mfs_strerr(error));
        goto errout;
    }
    MVFS_SET_VROOT(rtvp);

    /* Add "." and "..".  We need an extra hold for each one.  Log any errors,
    ** but continue on.
    */
    VN_HOLD(rtvp);
    if ((error = mfs_ramdir_add(rtvp, ".", rtvp, &num, cd)) != 0) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_viewroot_common_init: adding . err=%d%s...Continuing\n",
                 error, mfs_strerr(error));
        error = 0;
    }
    VN_HOLD(rtvp);
    if ((error = mfs_ramdir_add(rtvp, "..", rtvp, &num, cd)) != 0) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_viewroot_common_init: adding .. err=%d%s...Continuing\n",
                 error, mfs_strerr(error));
        error = 0;
    }
    MVFS_INCREMENT_LINK(rtvp);

    /* Add the ".specdev" object for ioctl's.  Each type of viewroot (currently
    ** automatic or dynamic) gets its own specdev (see the comment in
    ** mfs_ioctl.h).
    */
    if ((error = mfs_makespecnode(MFS_SDEVCLAS, (VNODE_T *)NULL,
                                  (void *)NULL, vfsp, &svp, cd)) != 0)
    {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_viewroot_common_init: make specdev for %s viewroot=%p"
                 " err=%d%s...Continuing\n",
                 MVFS_VWROOT_TO_STR(vrtype), rtvp, error, mfs_strerr(error));
        error = 0;
    } else {
        /* Use makespecnode hold count, no extra hold needed.  We use the same
        ** name for the .specdev in every viewroot (there's no reason not to).
        */
        if ((error = mfs_ramdir_add(rtvp, MVFS_SPECDEV, svp, &num, cd)) != 0) {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_viewroot_common_init: add specdev for %s"
                     " viewroot=%p type=%d err=%d%s...Continuing\n",
                     MVFS_VWROOT_TO_STR(vrtype), rtvp, mmip->mmi_mnttype,
                     error, mfs_strerr(error));
            error = 0;
        }
    }
    /* Now set up the root vnode in the mount point. */
    mmip->mmi_rootvp = rtvp;

    /* Save info in the viewroot structure (no extra hold counts on the rtvp and
    ** svp are needed).
    */
    vrdp->mfs_viewroot_vfsp = vfsp;
    vrdp->mfs_viewroot_vp = rtvp;
    vrdp->mfs_viewroot_specvp = svp;

  errout:
    return(error);
}

/*
 * Mfs mount in a useful internal form.
 * Set up mount info record and attach it to vfs struct.
 */
int
mvfs_vmount_subr(
    VFS_T *vfsp,    /* Our vfs ptr (mount tab) */
    VNODE_T *vwp,   /* if NULL, viewroot/global VOB, else view vp for au VOB. */
    char *path,     /* Pathname of special file */
    int flags,      /* flags to the mount */
    caddr_t data,   /* data args */
    size_t datalen, /* data args length */
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo /* Caller info (irp info)    */
)
{
    int error;               /* error number */
    VNODE_T *rtvp;           /* the server's root */
    struct mfs_mnode *mnp;   /* root MFS vinfo */
    struct mfs_mntinfo *mmi; /* MFS mount info */
    register int i;
    char *pnp;
    int vobminor, minor_num;
    dev_t major_num;
    int nvminor_num;
    dev_t nvmajor_num;
    tbs_boolean_t saved_largeinit = FALSE;
    tbs_boolean_t mvfs_misc_init_done = FALSE;
    VFS_T *canary;
    /* Declare a type so we can do one allocation to save stack space. */
    struct {
        struct mfs_mntargs mma;
        struct mvfs_splitpool splitpool;
        u_long cookie;
        struct mvfs_splitpool *splitpoolp;
    } *alloc_unitp;

    struct mfs_mntargs *mmap;
    mvfs_common_data_t *mcdp;
    mvfs_vfs_data_t *mvdp;
    mvfs_viewroot_data_t *vrdp;
    ks_uint32_t lstate;

    /* Before we get any locks, do a quick sanity check, did we get
     * any data passed to us?
     */
    if ((data == NULL) || (datalen <= 0)) {
        return(EINVAL);
    }

    MDKI_ATOMIC_INCR_UINT32(&mvfs_us_ops_in_progress);

    lstate = MDKI_ATOMIC_READ_UINT32(&mvfs_state);

    if (lstate == MVFS_UNLOAD_IN_PROGRESS ||
        lstate == MVFS_UNINITIALIZED)
    {
        MDKI_ATOMIC_DECR_UINT32(&mvfs_us_ops_in_progress);
        return(EIO);
    }

    /* Allocate stuff to keep the stack size down. */
    if ((alloc_unitp = KMEM_ALLOC(sizeof(*alloc_unitp), KM_SLEEP)) == NULL) {
        MDKI_ATOMIC_DECR_UINT32(&mvfs_us_ops_in_progress);
        return(ENOMEM);
    }

    mmap = &(alloc_unitp->mma);
    alloc_unitp->splitpoolp = &(alloc_unitp->splitpool);
    alloc_unitp->cookie = 0;

    mcdp = MDKI_COMMON_GET_DATAP();

    /* Find the right VFS data for this call.  NULL means it is a viewroot or a
    ** global VOB mount (for a dynamic view or an NT view), and non-NULL means
    ** it is a VOB mount for the specified automatic view.
    */
    mvdp = MDKI_VFS_GET_DATAP(vwp);

    ASSERT(ISLOCKEDBYME(&(mvdp->mount_lock)));

    error = 0;
    mmi = NULL;
    rtvp = NULL;

    /* Get the mount args */
    if ((error = CopyInMfs_mntargs(data, mmap, callinfo)) != 0) {
        goto errout;
    }
    /* Check the mount args version.  */
    if (mmap->mma_mntvers != MFSMNT_VERSION) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_vmount_subr: version mismatch: expected %d got %d\n",
                 MFSMNT_VERSION, mmap->mma_mntvers);
        error = EINVAL;
        goto errout;
    }
    if (mmap->mma_mntsize != sizeof(*mmap)) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_vmount_subr: parameter size mismatch"
                 " (got %d wanted %d)\n",
                 mmap->mma_mntsize, sizeof(*mmap));
        error = EINVAL;
        goto errout;
    }
    /* Do initialization and checking based on whether this is a viewroot
    ** (automatic or dynamic) or VOB mount and whether it is the very first
    ** mount or not.
    */
    if (MVFSMNT_IS_VIEWROOT(mmap)) {
        /* Viewroots better have vwp == NULL (or mvdp would be bad, i.e., not
        ** the global one).
        */
        ASSERT(vwp == NULL);

        if (mvdp->mount_count == 0) {
            /* Complete remaining phases during first mount of this zone.  This
            ** could be either the automatic or the dynamic viewroot mount (if
            ** they both get mounted, only the first one will do this final
            ** setup).
            */
            if (mmap->mma_sizes.version != MVFS_SETCACHE_VERSION) {
                mvfs_log(MFS_LOG_ERR,
                         "mvfs_vmount_subr: cache size version mismatch"
                         " (got %d wanted %d)\n",
                         mmap->mma_sizes.version, MVFS_SETCACHE_VERSION);
                error = EINVAL;
                goto errout;
            }
            saved_largeinit = TRUE;
            mcdp->mvfs_init_sizes.size[MVFS_SETCACHE_LARGEINIT] =
                mcdp->mvfs_largeinit;

            if ((error = mvfs_misc_init(&mmap->mma_sizes)) != 0) {
                goto errout;
            }
            mvfs_misc_init_done = TRUE;

            /* We need to finish initializing the vfs data structure here now
             * we've setup the tunables needed to do so.
             */
            error = mvfs_vfs_data_init(MVFS_VDIF_VOB_MOUNTS,
                                       mvdp,
                                       NULL);
            if (error != 0) {
                goto errout;
            }
#ifndef MVFS_DEBUG
            /*
             * done with startup messages, crank down logpri to normal
             * starting value.
             */
            mfs_logpri = MFS_LOG_ERR;
#endif
        }
    } else {
        /* This must be a VOB mount, so do some checking.  Only check for
         * duplicate mounts after everything is initialized in a given zone
         * (verified by checking mvdp->mount_count != 0).  Note, we may
         * need to allocate a minormap.
         */
        ASSERT((mmap->mma_flags & MFSMNT_VOB) &&
               ((vwp == NULL) || (MVFS_VIEW_IS_AUTO_VIEW(vwp))));

        /* Make sure at least one viewroot has been mounted because that's how
        ** we finish our initialization before we can mount VOBs.  The viewroots
        ** are counted in the "global" vfs data, and we might be doing an
        ** automatic view VOB mount (which points at per-view vfs data) so we
        ** need to make sure we use the global vfs data pointer here.
        */
        if (MDKI_VFS_GET_DATAP(NULL)->mount_count == 0) {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_vmount_subr: a viewroot (automatic or dynamic) must"
                     " be mounted before VOBs are mounted (mma_flags=0x%lx)\n",
                     mmap->mma_flags);
            error = EIO;
            goto errout;
        }
        while ((canary = mvfs_findvfs_cookie(mvdp, &(alloc_unitp->cookie)))
               != NULL)
        {
            if (canary == vfsp) {
                mvfs_log(MFS_LOG_ERR,
                         "mvfs_vmount_subr: duplicate vfsp %lx already mounted"
                         " at ID %lu\n",
                         vfsp, alloc_unitp->cookie);
                /* Should we mark the existing vfsp mount as broken somehow? */
                ASSERT(canary != vfsp); /* die on a debug kernel */
                error = EBUSY;
                goto errout;            /* fail on non-debug kernel */
            }
        }
    }
    /* We're back to "common" code (same code for viewroot and VOB mounts). */
#ifdef MVFS_HAS_DELAYED_PROCESSING
    /*
     * At this point, the mvfs_misc_init routine was already called and
     * it is safe to get the current mvfs_thread_t structure if necessary.
     * It is only in the delayed processing case that we care about the
     * thread structure.
     */
    if (cd->thr == NULL)
        cd->thr = mvfs_mythread(NULL);
#endif

    /* After this point, any error should cleanup everything we've done above */

    /* stats structures are not available till after misc_init */
    BUMPSTAT(mfs_vfsopcnt[MFS_VMOUNT]);

    /* Allocate room for the mount info struct */
    mmi =
        (struct mfs_mntinfo *)KMEM_ALLOC(sizeof(struct mfs_mntinfo), KM_SLEEP);
    if (mmi == NULL) {
        error = ENOMEM;
        goto errout;
    }
    BZERO(mmi, sizeof(struct mfs_mntinfo));
    mmi->mmi_minor = mmi->mmi_nvminor = -1; /* mark unallocated */
    INITSPLOCK(mmi->mmi_rclock, "mvfs_mmi_refcnt"); /* Initialize refcnt lock */
    mmi->mmi_zoneid = MDKI_GETZONEID();

    /* Set type of mount */
    if (MVFSMNT_IS_VIEWROOT(mmap)) {
        if (mmap->mma_vobminor != 0) {
            error = EINVAL;
            goto errout;
        }
        if (mmap->mma_flags & MFSMNT_VIEWROOT) {
            mmi->mmi_mnttype = MFS_VIEWSVRMNT; /* Dynamic viewroot. */
        } else {
            ASSERT(mmap->mma_flags & MVFSMNT_RVIEWROOT);
            mmi->mmi_mnttype = MVFS_RVIEWSVRMNT; /* Automatic viewroot. */
        }
    } else {
        /* It's a VOB mount. */
        if ((MDKI_INGLOBALZONE()) && (mmap->mma_vobminor != 0)) {
            /*
             * User space requests 1,2, ..., MVFS_MAJFIXMAX * MVFS_MINORMAX / 2.
             * we convert this to:
             * (MVFS_MAJDYNMAX + MVFS_MAJFIXMAX) * MVFS_MINORMAX - 2,
             * (MVFS_MAJDYNMAX + MVFS_MAJFIXMAX) * MVFS_MINORMAX - 4,
             * (MVFS_MAJDYNMAX + MVFS_MAJFIXMAX) * MVFS_MINORMAX - 6,
             * ...,
             * MVFS_MAJDYNMAX * MVFS_MINORMAX.
             *
             * The ncaexported vobs map to the high-addr part in the
             * mvfs_minormap, and going downward. But, the non-exported vobs
             * look for empty slots start from lower-addr part.
             *
             * Note: mma_vobminor is unsigned so math here is OK.
             */
            vobminor = (MVFS_MAJDYNMAX + MVFS_MAJFIXMAX) * MVFS_MINORMAX -
                2 * mmap->mma_vobminor;
            if (vobminor > (MVFS_MAJDYNMAX + MVFS_MAJFIXMAX) * MVFS_MINORMAX -
                2 || vobminor < MVFS_MAJDYNMAX * MVFS_MINORMAX)
            {
                error = EINVAL;
                goto errout;
            }
            if (mvfs_get_minorpair(&mvfs_minormap,
                                   MVFS_MINMAPSIZE,
                                   vobminor) != 0)
            {
                error = EBUSY; /* FIXME: better code for device in use ?*/
                goto errout;
            }
            mmi->mmi_minor = vobminor;
            mmi->mmi_nvminor = vobminor + 1;
        }

        mmi->mmi_mnttype = MFS_VOBMNT;

        /* If this is an automatic VOB mount, hold the view vnode.  Otherwise,
        ** it is a viewroot or dynamic VOB mount and vwp is NULL (we checked
        ** above).  In either case, put it in the mount structure for use later
        ** on.
        */
        if (vwp != NULL) {
            VN_HOLD(vwp);
        }
        mmi->mmi_viewvp = vwp;
    }
    /* For all mount types (viewroots or VOBs) fill in common information. */

    /* Fill in common "server" and "retry" structs.
     *     For Viewroot this is albd server (dynamic view) or rview agent
     *         (automatic view) info
     *     For Vobs this is vob svr/retry and view retry info.
     *
     * XXXODFS Figure out what this stuff means for automatic viewroots,
     * especially the mmi_svr stuff below.  There's probably no host, so don't
     * even call mfs_copyin_strbuf (which returns EINVAL if the str is NULL) for
     * automatic viewroots.
     */
    if (!MVFS_ISRVIEWSVRMNT(mmi) &&
        ((error = mfs_copyin_strbuf(mmap->mma_host, &mmi->mmi_svr.host)) != 0))
    {
        goto errout;
    }
    mmi->mmi_retry.soft = ((mmap->mma_flags & MFSMNT_SOFT) != 0);
    mmi->mmi_retry.nointr = ((mmap->mma_flags & MFSMNT_NOINTR) != 0);
    mmi->mmi_retry.timeo = mmap->mma_timeo;
    mmi->mmi_retry.retries = mmap->mma_retries;

    if (mmi->mmi_retry.timeo == 0) {
        /* By the time we get here, we've already filled in mmi->mmi_mnttype.
         * So, use that to figure out the default timeout value to use.
         */
        switch (mmi->mmi_mnttype) {
          case MFS_VIEWSVRMNT:
            mmi->mmi_retry.timeo = MFSMNT_TIMEO_DEFAULT;
            break;
          case MVFS_RVIEWSVRMNT:
            mmi->mmi_retry.timeo = MFSMNT_AU_TIMEO_DEFAULT;
            break;
          case MFS_VOBMNT:
            /*  If vwp is not NULL, this must be an auto view VOB mount. */
            if (vwp != NULL) {
                mmi->mmi_retry.timeo = MFSMNT_AU_TIMEO_DEFAULT;
            } else {
                mmi->mmi_retry.timeo = MFSMNT_TIMEO_DEFAULT;
            }
            break;
          default:
            mmi->mmi_retry.timeo = MFSMNT_TIMEO_DEFAULT;
            break;
        }
    }

    if (mmi->mmi_retry.retries == 0)
        mmi->mmi_retry.retries = MFSMNT_RETRANS_DEFAULT;

    /* Fill mount dir pathname */

    error = mfs_copyin_strbuf(mmap->mma_mntpath, &mmi->mmi_mntpath);
    if (error) goto errout;

    /* Now do mount type specific setup.  Either type of viewroot could be
    ** mounted first.
    */
    if (mmi->mmi_mnttype == MFS_VIEWSVRMNT) {
        if ((error = mvfs_viewroot_common_init(vfsp, DYVWROOT, mmap, mmi, cd))
            != 0)
        {
            goto errout;
        }
       /* Fill in viewroot server struct */
        pnp = STRDUP("albd server");
        MFS_INIT_STRBUFPN_PAIR_IN(&mmi->mmi_svr.lpn, pnp, pnp);
        mmi->mmi_svr.rpn = STRDUP("albd server");
        mmi->mmi_svr.uuid = TBS_UUID_NULL;
        mmi->mmi_svr.svrbound = 1;

        /* We only care about the ALBD port for the viewroot mount.  The port
        ** is assumed to be the same for IPv6 and IPv4 so we're just going to
        ** "misuse" the addr field in this case and say it is IPv4 and save the
        ** port in it (we set the family for safety).
        */
        mmi->mmi_svr.addr.ks_ss_s.sa_family = AF_INET;
        mmi->mmi_svr.addr.ks_ss_sin4.sin_port = mmap->mma_port;

        goto out;
    } else if (mmi->mmi_mnttype == MVFS_RVIEWSVRMNT) {
        if ((error = mvfs_viewroot_common_init(vfsp, AUVWROOT, mmap, mmi, cd))
            != 0)
        {
            goto errout;
        }

        /* Automatic views are like Windows views on all platforms. */
        mmi->mmi_default_vwtag_kind = MFS_NTVWCLAS;

        goto out;
    } else {
        /* Just fall through to save an indent level for VOB mounts. */
    }
    /*
     * VOB MOUNT - fill in fields only associated with VOB mounts.  This could
     * be a Unix/Linux VOB mount (called through the file system) or a Windows
     * VOB mount (called by the MVFS_CMD_MOUNT ioctl).  We require that the
     * right type of viewroot be already mounted (dynamic or automatic).
     */

    /* The only use of vrdp is to get the NOSUID flag from the mount.  We can
    ** tell which viewroot based on the vwp being NULL (dynamic view) or
    ** non-NULL (automatic view).
    */
    if (vwp != NULL) {
        ASSERT(MFS_VPISMFS(vwp) && MVFS_VIEW_IS_AUTO_VIEW(vwp));
        vrdp = MDKI_VIEWROOT_GET_DATAP(AUVWROOT);
    } else {
        vrdp = MDKI_VIEWROOT_GET_DATAP(DYVWROOT);
    }
    if (vrdp->mfs_viewroot_vfsp == NULL) {
        /* viewroot not mounted--this will screw up minor number allocation
           policy for exported views.  We must error out. */
        /* FIXME:  what error code? */
        error = EINVAL;
        goto errout;
    }
    /*
     * Fill in vob-tag name.
     * NOTE:  this is a pointer in the mntpath name, and so
     * is not independently freed in unmount.
     */

    if (mmi->mmi_mntpath) {
        mmi->mmi_vobtag = MVFS_FIND_LEAF(mmi->mmi_mntpath);
        /*
         * Don't allow duplicate MVFS mounts at the same "pathname"
         * On Unix
         *  - While "mount on mount" might theoretically work, I think
         *    it is likely to cause trouble, and certainly confusing to have
         *    to think about all the time... so its good to be sure it
         *    isn't happening.  (Think about looking up ".." at a vob root
         *    that has 2 or more vob roots mounted at the same point to make
         *    your head hurt).
         */
        if (mvfs_findvfs_mntpath(mvdp, mmi->mmi_mntpath) != NULL) {
            error = EEXIST;
            goto errout;
        }
    } else {
        mmi->mmi_vobtag = NULL;
    }

    /* VOB MOUNT - Fill in vob-only portions of mmi_svr struct */

    MFS_STRBUFPN_PAIR_GET_UPN(&mmi->mmi_svr.lpn).s = NULL;
    MFS_STRBUFPN_PAIR_GET_UPN(&mmi->mmi_svr.lpn).l = 0;
    MFS_STRBUFPN_PAIR_GET_UPN(&mmi->mmi_svr.lpn).m = 0;
    error =
        mfs_copyin_strbufpn(MFS_STRBUFPN_PAIR_GET_KPN(&mmap->mma_spath),
                            &MFS_STRBUFPN_PAIR_GET_KPN(&mmi->mmi_svr.lpn).s);
    if (error) goto errout;

    MFS_STRBUFPN_PAIR_GET_KPN(&mmi->mmi_svr.lpn).l =
        STRLEN(MFS_STRBUFPN_PAIR_GET_KPN(&mmi->mmi_svr.lpn).s);
    MFS_STRBUFPN_PAIR_GET_KPN(&mmi->mmi_svr.lpn).m =
        MFS_STRBUFPN_PAIR_GET_KPN(&mmi->mmi_svr.lpn).l + 1;
    if ((MFS_STRBUFPN_PAIR_GET_UPN(&mmap->mma_spath).s != NULL) &&
                (MFS_STRBUFPN_PAIR_GET_UPN(&mmap->mma_spath).s !=
                 MFS_STRBUFPN_PAIR_GET_KPN(&mmap->mma_spath).s))
    {
        error =
           mfs_copyin_strbufpn(MFS_STRBUFPN_PAIR_GET_UPN(&mmap->mma_spath),
                               &MFS_STRBUFPN_PAIR_GET_UPN(&mmi->mmi_svr.lpn).s);
        if (error) goto errout;
        MFS_STRBUFPN_PAIR_GET_UPN(&mmi->mmi_svr.lpn).l =
            STRLEN(MFS_STRBUFPN_PAIR_GET_UPN(&mmi->mmi_svr.lpn).s);
        MFS_STRBUFPN_PAIR_GET_UPN(&mmi->mmi_svr.lpn).m =
            MFS_STRBUFPN_PAIR_GET_UPN(&mmi->mmi_svr.lpn).l + 1;
    }
    error = mfs_copyin_strbuf(mmap->mma_rpath, &mmi->mmi_svr.rpn);
    if (error) goto errout;
    mmi->mmi_svr.uuid = mmap->mma_replica_uuid;

    if (mmap->mma_sptab_cnt) {
        mvfs_log(MFS_LOG_DEBUG, "copyin sptab cnt %d\n", mmap->mma_sptab_cnt);
        /*
         * copy in, translate, and verify the format of pathname
         * translation items.
         */
        mmi->mmi_sptable_len = mmap->mma_sptab_cnt;
        mmi->mmi_sptable = (mvfs_sp_ent_t *)
            KMEM_ALLOC(mmi->mmi_sptable_len * sizeof(*mmi->mmi_sptable),
                       KM_SLEEP);
        if (mmi->mmi_sptable == NULL) {
            error = ENOMEM;
            goto errout;
        }
        BZERO(mmi->mmi_sptable,
              mmi->mmi_sptable_len * sizeof(*mmi->mmi_sptable));
        for (i = 0; i < mmap->mma_sptab_cnt; i++) {
            error = CopyInMvfs_splitpool_index((caddr_t)mmap->mma_sptab_ents,
                                               alloc_unitp->splitpoolp,
                                               i,
                                               callinfo);

            if (error)
                goto errout;
            error = mfs_copyin_strbuf((alloc_unitp->splitpoolp)->msp_prefix,
                                      &mmi->mmi_sptable[i].sp_prefix);
            if (error)
                goto errout;
            mmi->mmi_sptable[i].sp_prlen =
                STRLEN(mmi->mmi_sptable[i].sp_prefix);
            if (mmi->mmi_sptable[i].sp_prefix[mmi->mmi_sptable[i].sp_prlen-1]
                != MVFS_PN_SEP_CHAR)
            {
                error = EINVAL; /* prefix must end in sep char */
                goto errout;
            }
            /* Copy in kernel and user versions of the replacement */
            error =
                mfs_copyin_strbufpn(
                    MFS_STRBUFPN_PAIR_GET_KPN(
                        &((alloc_unitp->splitpoolp)->msp_target)),
                    &mmi->mmi_sptable[i].sp_target);
            if (error)
                goto errout;
            mmi->mmi_sptable[i].sp_tglen =
                STRLEN(mmi->mmi_sptable[i].sp_target);
            error =
                mfs_copyin_strbufpn(
                    MFS_STRBUFPN_PAIR_GET_UPN(
                        &((alloc_unitp->splitpoolp)->msp_target)),
                    &mmi->mmi_sptable[i].sp_usertarget);
            if (error)
                goto errout;
            mmi->mmi_sptable[i].sp_usrlen =
                STRLEN(mmi->mmi_sptable[i].sp_usertarget);
            if (mmi->mmi_sptable[i].sp_target[mmi->mmi_sptable[i].sp_tglen-1]
                != MVFS_PN_SEP_CHAR ||
                mmi->mmi_sptable[i].sp_usertarget[mmi->mmi_sptable[i].sp_usrlen-1]
                != MVFS_PN_SEP_CHAR)
            {
                error = EINVAL; /* replacement must end in sep char */
                goto errout;
            }
        }
    }
    /* Fill in vob-only vob-oid args */

    mmi->mmi_voboid = mmap->mma_vob_oid;
    mmi->mmi_vobuuid = mmap->mma_replica_uuid;

    /* Fill in other vob-mount paramters (e.g. caches) */

    mmi->mmi_noac = ((mmap->mma_flags & MFSMNT_NOAC) != 0);
    mmi->mmi_nodnlc = ((mmap->mma_flags & MFSMNT_NODNLC) != 0);
    mmi->mmi_ac_regmax = mmap->mma_ac_regmax;
    mmi->mmi_ac_regmin = mmap->mma_ac_regmin;
    mmi->mmi_ac_dirmax = mmap->mma_ac_dirmax;
    mmi->mmi_ac_dirmin = mmap->mma_ac_dirmin;

    /* Other defaults if not set in mount args */

    if (mmi->mmi_ac_regmax == 0) mmi->mmi_ac_regmax = 60;
    if (mmi->mmi_ac_regmin == 0) mmi->mmi_ac_regmin = 3;
    if (mmi->mmi_ac_dirmax == 0) mmi->mmi_ac_dirmax = 60;
    if (mmi->mmi_ac_dirmin == 0) mmi->mmi_ac_dirmin = 30;

    /* Initialize dir attribute cache flush time */
    mmi->mmi_ac_dir_ftime = 0;

    /* no need to lock mmi_refcnt here as noone else can be looking at it */
    mmi->mmi_refcnt = 0;
    if (mmi->mmi_minor == -1) {
        mmi->mmi_minor = mvfs_vfsgetnum(&mvfs_minormap, MVFS_MINMAPSIZE);
        if (mmi->mmi_minor < 0) {
            error = ENFILE;
            goto errout;
        }
        mmi->mmi_nvminor = mvfs_vfsgetnum(&mvfs_minormap, MVFS_MINMAPSIZE);
        if (mmi->mmi_nvminor < 0) {
            error = ENFILE;
            goto errout;
        }
    }
    if ((error = mvfs_find_devnum(mmi->mmi_nvminor, &major_num, &minor_num)) != 0)
        goto errout;
    MAKEFSID(mmi->mmi_nvfsid, major_num, minor_num);
    mmi->mmi_gen = mfs_mntgen;
    mmi->mmi_rdev = 0;

    /* Fill in the vfs struct for the MFS mount */

    vfsp->vfs_bsize = DEV_BSIZE;
    vfsp->vfs_flag |= (vrdp->mfs_viewroot_vfsp->vfs_flag & MVFS_VFS_NOSUID);
    if ((mmap->mma_flags & MFSMNT_RDONLY) != 0)
        vfsp->vfs_flag |= VFS_RDONLY;

    MVFS_SET_FSTYPE(vfsp);
    MVFS_ATTACH_VFS_DATA(vfsp, mmi);
    if ((error = mvfs_find_devnum(mmi->mmi_minor, &major_num, &minor_num)) != 0)
        goto errout;
    MAKEFSID(VFS_FSID(vfsp), major_num, minor_num);
    if ((error = mvfs_find_devnum(mmi->mmi_nvminor, &nvmajor_num, &nvminor_num)) != 0)
        goto errout;
    /* nvminor is returned to VOP_GETATTR() for view-less vob root;
       use it for what we put in the vfs_dev slot (but let the mdep
       implementation choose if it needs to put something else there,
       e.g. Linux). */
    MAKEVFSDEV(vfsp, major_num, minor_num, nvmajor_num, nvminor_num);

    /* The root vnode for the MFS mount is made with
     * a canned fid (inumber).  The actual vob object
     * which the mount point represents may be different
     * in different views (a different version of the dir).
     * As a result, the actual binding to the correct vob/view
     * object can not be done until the user references the mount
     * point w/ a view context.
     */

    error = mfs_makevobrtnode(vwp, vfsp, &rtvp, cd);
    if (error) goto errout;

    /* We have made an inode.  Nothing must allow us to fail now
       or we will orphan an inode in the inode cache with a garbage
       back pointer to a mount point. */

    /* Now set up the flags and root vnode in the mount point. */

    MVFS_SET_VROOT(rtvp);
    mmi->mmi_rootvp = rtvp; /* MFS root vnode */

    /*
     * Add this VFS to the table of vob mounts.
     */
    mvdp->vob_mounts[mmi->mmi_minor] = vfsp;
    mvdp->vob_mount_hwm = (mmi->mmi_minor > mvdp->vob_mount_hwm) ?
                    mmi->mmi_minor : mvdp->vob_mount_hwm;

    /* Cleanup (if error) and return */

errout:
out:
    MDB_VFSLOG((MFS_VMOUNT,
                "rtvp=%"KS_FMT_PTR_T" devpath=%"KS_FMT_PTR_T", mntpath=%s, "
                "vfsp=%"KS_FMT_PTR_T", flags=%lx, err=%d\n",
                rtvp, path,
                (mmi && mmi->mmi_mntpath) ? mmi->mmi_mntpath : "", vfsp,
                vfsp->vfs_flag, error));

    if (error) {
        if (mmi) {
            FREESPLOCK(mmi->mmi_rclock);
            mfs_svrdestroy(&mmi->mmi_svr);
            if (mmi->mmi_mntpath) STRFREE(mmi->mmi_mntpath);
            if (mmi->mmi_hmsuffix) STRFREE(mmi->mmi_hmsuffix);
            if (mmi->mmi_hmvers_nm) STRFREE(mmi->mmi_hmvers_nm);
            if (mmi->mmi_sptable) {
                for (i = 0; i < mmi->mmi_sptable_len; i++) {
                    if (mmi->mmi_sptable[i].sp_prefix)
                        STRFREE(mmi->mmi_sptable[i].sp_prefix);
                    if (mmi->mmi_sptable[i].sp_target)
                        STRFREE(mmi->mmi_sptable[i].sp_target);
                    if (mmi->mmi_sptable[i].sp_usertarget)
                        STRFREE(mmi->mmi_sptable[i].sp_usertarget);
                }

                KMEM_FREE(mmi->mmi_sptable,
                          mmi->mmi_sptable_len * sizeof(*mmi->mmi_sptable));
            }
            if (mmi->mmi_minor >= 0)
                mvfs_vfsputnum(&mvfs_minormap,
                               MVFS_MINMAPSIZE,
                               mmi->mmi_minor);
            if (mmi->mmi_nvminor >= 0)
                mvfs_vfsputnum(&mvfs_minormap,
                               MVFS_MINMAPSIZE,
                               mmi->mmi_nvminor);

            if (mmi->mmi_viewvp != NULL) {
                ATRIA_VN_RELE(mmi->mmi_viewvp, cd);
            }

            KMEM_FREE(mmi, sizeof(struct mfs_mntinfo));
        }
        /* Must never get here with a rt vnode ptr.  The iput
         * to release it would not destroy it, just leave it cached.
         * I need a call that will really destroy the vnode.
         */
        ASSERT(rtvp == NULL);

        if (saved_largeinit)
            mcdp->mvfs_largeinit =
                mcdp->mvfs_init_sizes.size[MVFS_SETCACHE_LARGEINIT];

        if (mvfs_misc_init_done) {
            mvfs_misc_free();
        }
    } else {
        /* Everything is OK so record this mount. */

        /* There better be at least one mounted viewroot at this point. */
        ASSERT((MDKI_VIEWROOT_GET_DATAP(DYVWROOT)->mfs_viewroot_vfsp != NULL)
               ||
               (MDKI_VIEWROOT_GET_DATAP(AUVWROOT)->mfs_viewroot_vfsp != NULL));

        /* We depend on the global vfs data count going from 0 to 1 to know that
        ** everything is initialized.  If this is an automatic view, we can just
        ** count it.
        */
        if (vwp == NULL) {
            if (mvdp->mount_count++ == 0) {
                MDKI_ATOMIC_SET_UINT32(&mvfs_state, MVFS_FULLY_INITIALIZED);
            }
        } else {
            mvdp->mount_count++;
        }
    }

    /* We're relying on holding the mount_lock to fix mount/mount and
     * mount/unmount races. It should be ok to hold this lock across these calls
     * because none of the call paths attempt to lock the mount lock.
     * Additionally, the mount lock is freed in mvfs_unload.
     *
     * Only do this for the global vfs data structure...
     */
    if (vwp == NULL && mvdp->mount_count == 0) {
        /* This is the error case since the mount_count is still 0 (which
        ** means the mount we were trying to do must not have worked) so
        ** cleanup.
        */
        if (mvfs_misc_init_done) {
            mvfs_misc_free();
        }

#ifndef MVFS_DEBUG
        /* return logpri to default value */
        mfs_logpri = MFS_LOG_INFO;
#endif
    }

    MDKI_ATOMIC_DECR_UINT32(&mvfs_us_ops_in_progress);
    /* Note, freeing the struct mfs_mntargs here doesn't leak memory because
    ** all the strings that were copyin'd and pointed to by mma_xxx fields
    ** (e.g. mma_host) are also pointed to by mmi_xxx fields, so they will be
    ** freed when the struct mfs_mntinfo is (or if we had an error above).
    */
    KMEM_FREE(alloc_unitp, sizeof(*alloc_unitp));
    return (error);
}

/*
 * mfs_vunmount vfsop
 */

int
mvfs_vunmount(
    VFS_T *vfsp,
    CALL_DATA_T *cd
)
{
    mvfs_vfs_data_t *mvdp = MDKI_VFS_GET_DATAP(VFS_TO_MMI(vfsp)->mmi_viewvp);
    int error;

    MVFS_LOCK(&(mvdp->mount_lock));

    error = mvfs_vunmount_subr(vfsp, cd);

    MVFS_UNLOCK(&(mvdp->mount_lock));

    return error;
}

int
mvfs_vunmount_subr(
    VFS_T *vfsp,
    CALL_DATA_T *cd /* Cred may be NULL from pre-SVR4 wrappers */
)
{
    VNODE_T *rtvp; /* Root vnode ptr */
    struct mfs_mntinfo *mmi; /* Mount info */
    int error = 0;
    int count;
    int refcnt;
    SPL_T s;
    mvfs_vfs_data_t *mvdp = MDKI_VFS_GET_DATAP(VFS_TO_MMI(vfsp)->mmi_viewvp);
    mvfs_viewroot_data_t *vrdp;
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();
    mvfs_viewroot_type_t vrtype;

    MDKI_ATOMIC_INCR_UINT32(&mvfs_us_ops_in_progress);

    switch (MDKI_ATOMIC_READ_UINT32(&mvfs_state)) {
      case MVFS_UNINITIALIZED:
      case MVFS_UNLOAD_IN_PROGRESS:
      case MVFS_PHASE_1_INITIALIZED:
        MDKI_ATOMIC_DECR_UINT32(&mvfs_us_ops_in_progress);
        return(EIO);
      default:
        break;
    }

    BUMPSTAT(mfs_vfsopcnt[MFS_VUMOUNT]);

    mmi = VFS_TO_MMI(vfsp);

    ASSERT(mmi->mmi_rootvp != NULL);

    vrtype = MVFS_VRTYPE_VFS(vfsp);
    vrdp = MDKI_VIEWROOT_GET_DATAP(vrtype);

    MDB_VFSLOG((MFS_VUMOUNT,
                "vfsp=%"KS_FMT_PTR_T" rtvp=%"KS_FMT_PTR_T" mntpath=%s\n",
                vfsp, mmi->mmi_rootvp, mmi->mmi_mntpath));

    /*
     * The refcnt check on the viewsvr mount is relatively
     * complicated.  We must first uncache all hold counts
     * on views (from any vob, the name cache, or loopback
     * objects), and then check that every view has
     * exactly 1 refcount left on it, and the total
     * refcnts on the mount is only the contents of the /view dir.
     * If it passes these checks, then the unmount should be
     * able to succeed and so it is OK to destroy the contents
     * of the /view dir.
     */

    if (MVFS_ISVIEWROOTMNT(mmi)) {
        /* This flushes everything for all types of viewroots.  This normally
        ** only happens when we're shutting down Clearcase (and unmounting the
        ** viewroots) so there's not much reason to restrict the flushing to be
        ** more "fine grained".  We pass in the vrtype in case this changes in
        ** the future.
        */
        mvfs_dncflush(vrtype, cd); /* Flush name cache - it HOLDS views */
        mvfs_mnflush(vrtype, cd);  /* Flush all vobs, viewsvr mount */

        /*
         * Check that each view has only 1 ref, and get total count
         * of objects.
         */

        error = mfs_ramdir_refchk(mmi->mmi_rootvp, &count);
        if (error) goto errout;
        SPLOCK(mmi->mmi_rclock, s);
        if (mmi->mmi_refcnt > count+1) { /* +1 ref on mount point */
            refcnt = mmi->mmi_refcnt;
            SPUNLOCK(mmi->mmi_rclock, s);
            mvfs_log(MFS_LOG_ERR,
                     "umount: ref check failed: refcnt=%d expected %d\n",
                     refcnt,
                     count + 3);
            error = EBUSY;
            goto errout;
        }
        SPUNLOCK(mmi->mmi_rclock, s);

        if (V_COUNT(mmi->mmi_rootvp) > MVFS_UMOUNT_EXPECTED_REFCHK) {
            mvfs_log(MFS_LOG_ERR,
                     "umount: ref check on viewrootvp %"KS_FMT_PTR_T" failed: "
                     "refcnt=%d expected %d\n",
                     mmi->mmi_rootvp,
                     V_COUNT(mmi->mmi_rootvp),
                     MVFS_UMOUNT_EXPECTED_REFCHK);
            error = EBUSY;
            goto errout;
        }

        /*
         * Passes all refcnt checks - now PURGE the contents
         * of the RAMDIR.
         */
        mfs_ramdir_purge(mmi->mmi_rootvp, RELEASE, cd);
    }

    /*
     * Flush the name cache.  This has VFS pointers in it
     * which may be invalid after this unmount.
     */

    mvfs_dnc_flushvfs(vfsp, cd);

    /* Invalidate cached mnodes */

    mfs_mnflushvfs(vfsp, cd); /* Uncache all idle vnodes */

    /* Make sure the mount point isn't busy. */

    SPLOCK(mmi->mmi_rclock, s);
    if (mmi->mmi_refcnt > 1) {
        refcnt = mmi->mmi_refcnt;
        SPUNLOCK(mmi->mmi_rclock, s);
        mvfs_log(MFS_LOG_ERR,
                 "umount: mmi ref check on %s failed: refcnt=%d expected 1\n",
                 mmi->mmi_mntpath,
                 refcnt);
        error = EBUSY;
        goto errout;
    }

    SPUNLOCK(mmi->mmi_rclock, s);

    rtvp = mmi->mmi_rootvp;
    if (V_COUNT(rtvp) != MVFS_UMOUNT_BUSY_REFCHK) {
        mvfs_log(MFS_LOG_ERR,
                 "umount: ref check on %s rootvp %"KS_FMT_PTR_T" failed: "
                 "refcnt=%d expected %d\n",
                 mmi->mmi_mntpath,
                 rtvp,
                 V_COUNT(rtvp),
                 MVFS_UMOUNT_BUSY_REFCHK);
        error = EBUSY;
        goto errout;
    }
    /* Now we start doing "serious damage" (setting pointers to NULL, rele'ing
    ** vnodes, etc.) from which we can't recover, so don't try to return an
    ** error after this point.
    **
    ** XXXODFS The mfs_ramdir_purge() call above seems to fall into this
    ** category, too, so I wonder if it shouldn't be moved down below here (just
    ** inside the next if below).  We'd have to check if there's any interaction
    ** with the flushes above and maybe even the refcnt checks.
    */
    /*
     * Free up and release mfs root inode.  Note that since VN_RELE
     * may decide to "cache" the vnode, we must force it out ourselves
     * by flushing the mnode cache again.
     */

    if (MVFS_ISVIEWROOTMNT(mmi)) {

        vrdp->mfs_viewroot_vfsp = NULL;
        vrdp->mfs_viewroot_vp = NULL;
        vrdp->mfs_viewroot_specvp = NULL;

        if (MDKI_INGLOBALZONE() && (mvdp->mount_count == 1)) {
            /* If this is the last viewroot to be unmounted, transition to
             * MVFS_PHASE_1_INITIALIZED here before we start tearing down
             * things. This prevents a possible race between this unmount and a
             * sync.
             *
             * It's important to note here that after this point this function
             * cannot fail as there is no safe way to restore mvfs_state.
             */
state_sw:
            switch (MDKI_ATOMIC_READ_UINT32(&mvfs_state)) {
              case MVFS_FULLY_INITIALIZED:
                if (MDKI_ATOMIC_CAS_UINT32(&mvfs_state,
                                           MVFS_FULLY_INITIALIZED,
                                           MVFS_PHASE_1_INITIALIZED))
                {
                    break;
                }
                /* Fall through if the switch failed. */
              case MVFS_SYNC_IN_PROGRESS:
                MDKI_USECDELAY(50000);
                goto state_sw;
              default:
                MDKI_PANIC("mvfs_vunmount: mvfs_state incorrect when "
                           "transitioning to MVFS_PHASE_1_INITIALIZED.");
            }

            /* Close the logfile here (when unmounting /view) since we may need
             * a viewroot in order to call some of the log file internal
             * functions (e.g. MVOP_WRITE_KERNEL()).  Since we might have
             * multiple viewroots, only close if there's only one left.
             */
            mvfs_logfile_close(cd);
        } else {
            /* mvfs_vsync uses the mfs_viewroot_vfsp without taking the
             * mount_lock so we need to wait here after setting
             * mfs_viewroot_vfsp to NULL if sync is running. Note here that
             * although this routine sets mfs_viewroot_vp to NULL, the root vp
             * is actually rele'd below (we stashed it in 'rtvp',
             * mfs_viewroot_vp and mmi_rootvp are the same).
             *
             * NOTE: This was done (eliminating mvfs_vsync taking the
             * mount_lock) to fix a possible deadlock where this unmount is
             * waiting for MVFS_SYNC_IN_PROGRESS to end while mvfs_vsync is
             * waiting for the mount_lock.
             */
            while (MDKI_ATOMIC_READ_UINT32(&mvfs_state)
                    == MVFS_SYNC_IN_PROGRESS)
            {
                MDKI_USECDELAY(50000);
            }
        }
    } else {
        /* This is a VOB mount. */
        mvdp->vob_mounts[mmi->mmi_minor] = NULL; /* Clear vfs table */
        /*
         * Some platforms (e.g. Linux) will need the core code to reduce the
         * root vnode's link count so that the inode is properly freed.  This is
         * taken care of in mfs_ramdir_purge for the view root.
         */
        MVFS_DECREMENT_LINK(rtvp);
    }

    ATRIA_VN_RELE(rtvp, cd);
    mmi->mmi_rootvp = NULL;

    /* Automatic view VOB mounts have an embedded pointer to the view that
     * contains them so release that here if there is one.
     */
    if (mmi->mmi_viewvp != NULL) {
        ATRIA_VN_RELE(mmi->mmi_viewvp, cd);
        mmi->mmi_viewvp = NULL;
    }

    mfs_mnflushvfs(vfsp, cd); /* Could be left in freelist */

    /* Release the minor device(s) allocated for this mount
       and increment the mount generation number so caches
       will not be confused if anyone reuses this minor
       dev in an fsid. */

    mvfs_vfsputnum(&mvfs_minormap, MVFS_MINMAPSIZE, mmi->mmi_minor);
    if (mmi->mmi_nvminor >= 0)
        mvfs_vfsputnum(&mvfs_minormap, MVFS_MINMAPSIZE, mmi->mmi_nvminor);

    /* Free up and release allocated structs and pointers */

    mfs_svrdestroy(&mmi->mmi_svr);
    if (mmi->mmi_mntpath) STRFREE(mmi->mmi_mntpath);
    if (mmi->mmi_hmsuffix) STRFREE(mmi->mmi_hmsuffix);
    if (mmi->mmi_hmvers_nm) STRFREE(mmi->mmi_hmvers_nm);
    if (mmi->mmi_sptable) {
        for (count = 0; count < mmi->mmi_sptable_len; count++) {
            if (mmi->mmi_sptable[count].sp_prefix)
                STRFREE(mmi->mmi_sptable[count].sp_prefix);
            if (mmi->mmi_sptable[count].sp_target)
                STRFREE(mmi->mmi_sptable[count].sp_target);
            if (mmi->mmi_sptable[count].sp_usertarget)
                STRFREE(mmi->mmi_sptable[count].sp_usertarget);
        }

        KMEM_FREE(mmi->mmi_sptable,
                  mmi->mmi_sptable_len * sizeof(*mmi->mmi_sptable));
    }

    /* Clean up mount table entry */

    MVFS_REMOVE_VFS_DATA(vfsp, mmi);
    FREESPLOCK(mmi->mmi_rclock);
    KMEM_FREE(mmi, sizeof(struct mfs_mntinfo));

    /* Always decrement the count.  If this is the last global unmount
    ** (presumably a viewroot), then clean stuff up (we don't have a mmi_viewvp
    ** anymore to tell if this is a global unmount or an auview unmount, but we
    ** used it to set mvdp so we can check that).
    */
    if (--(mvdp->mount_count) == 0 && mvdp == MDKI_VFS_GET_DATAP(NULL)) {
        mvfs_misc_free();
#ifndef MVFS_DEBUG
        /* return logpri to default value */
        mfs_logpri = MFS_LOG_INFO;
#endif
    }

errout:
    MDKI_ATOMIC_DECR_UINT32(&mvfs_us_ops_in_progress);

    return(error);
}

/*
 * Get file system statistics.
 * Note that statfs struct is a little different on FSS systems.
 */

int
mfs_vstatfs(
    register VFS_T *vfsp,
    STATVFS_T *sbp
)
{
    return(mvfs_vstatfs_common(vfsp, sbp, (CALL_DATA_T *)NULL));
}

int
mvfs_vstatfs_common(
    register VFS_T *vfsp,
    STATVFS_T *sbp,
    CALL_DATA_T *cd
)
{
    int error;
    struct mfs_mntinfo *mmi;
    CLR_VNODE_T *cvp;

    if (vfsp == NULL)
	return(EINVAL);

    BUMPSTAT(mfs_vfsopcnt[MFS_VSTATFS]);

    /* Do STATFS  for VOB/View mount. */

    mmi = VFS_TO_MMI(vfsp);

    MVFS_VSTATFS_BZERO(sbp, sizeof(*sbp));	/* Zero first ... */

    if (MVFS_ISVIEWROOTMNT(mmi)) {
        mvfs_viewroot_data_t *vrdp;

        vrdp = MDKI_VIEWROOT_GET_DATAP(MVFS_VRTYPE_VFS(vfsp));

	if (vfsp == vrdp->mfs_viewroot_vfsp) {
            MDB_VFSLOG((MFS_VSTATFS, "real viewroot, using ROOTDIR\n"));
	    error = MVFS_STATFS(ROOTDIR->v_vfsp, CLR_ROOTDIR, sbp, cd);
            STATVFS_RESET_FLAGS(sbp);
	    STATVFS_FILL(sbp);
	} else {
            /*
             * Just use ROOTDIR, this platform doesn't have cloned VFSes
             * that matter for statfs reasons.
             */
            MDB_VFSLOG((MFS_VSTATFS, "cloned viewroot %"KS_FMT_PTR_T", using ROOTDIR\n",
                        vfsp));
            error = MVFS_STATFS(ROOTDIR->v_vfsp, CLR_ROOTDIR, sbp, cd);
	    STATVFS_FILL(sbp);
        }

    } else {
        /* Not VIEWSVRMNT */
  	error =
            LOOKUP_STORAGE_FILE(TRUE,
                                MFS_STRBUFPN_PAIR_GET_KPN(&mmi->mmi_svr.lpn).s,
                                NULL, &cvp, cd);
	if (!error) {
	    error = MVFS_STATFS(MVFS_CVP_TO_VP(cvp)->v_vfsp, cvp, sbp, cd);
	    CVN_RELE(cvp, cd);
        }
	STATVFS_FILL(sbp);
    }

    MDB_VFSLOG((MFS_VSTATFS, "vfsp=%"KS_FMT_PTR_T" mtype=%d err=%d\n",
                vfsp, mmi->mmi_mnttype, error));
    return (error);
}

/*
 * MFS_ROOT - return held root for a vfs
 */

int
mfs_root(vfsp, vpp)
VFS_T *vfsp;
VNODE_T **vpp;
{
    struct mfs_mntinfo *mmi;
    int error;

    mmi = VFS_TO_MMI(vfsp);

    if (mmi) {
        *vpp = mmi->mmi_rootvp;
        error = 0;          /* Optimistically assume the normal case. */
        /*
        ** If we're at the view root (/view) then don't check the state of the
        ** view so that, e.g., /view/.specdev will work.  Do the view check for
        ** vob roots (/vobs/foo).
        */
        if (mmi->mmi_mnttype == MFS_VOBMNT) {
            VNODE_T *vw;
            /*
            ** Check if someone ended the view while we are in it.  This is a
            ** lookup of a vob root, so return ESTALE, just like we do for ".."
            ** in mfs_lookup() in this same situation.  Copy some code from
            ** mfs_bindroot() and mvfs_rvclookup() to check our situation.
	    **
            ** mfs_getview() can ignore the cred arg, and we don't have one
            ** anyway, so pass in NULL.
            */
            if ((vw = mfs_getview(*vpp, NULL /* cred */, FALSE /* NO HOLD */)) != NULL) {
                if (VTOM(vw)->mn_view.id == MFS_NULLVID) {
                    *vpp = NULL;
                    error = MVFS_ESTALE; /* View is gone. */
                }
            }
        }
        if (*vpp) {
            VN_HOLD(*vpp);      /* If we're returning a vnode, hold it. */
        }
    } else {
        *vpp = NULL;
	error = ENXIO;          /* No mount info. */
    }
    MDB_VFSLOG((MFS_VROOT, "vfsp=%"KS_FMT_PTR_T" rtvp= %"KS_FMT_PTR_T"\n", vfsp, *vpp));
    BUMPSTAT(mfs_vfsopcnt[MFS_VROOT]);
    return(error); 
}

/* Some state vars and defn for periodic maint/sync */

int mfs_do_sync = 1;
int mdb_crash = 0;

extern void mfs_prkmem(P_NONE);
/****************************************************************************
 * mfs_periodic_maintenance
 * Perform periodic maintenance tasks for the MVFS
 * No arguments
 * Returns:	no result
 *
 * Description:
 *	This routine performs the standard MVFS periodic maintenance
 *	that is not related to any mountpoint. The tasks this routine
 *	performs are:
 *	- Panics the system if the internal flag 'mdb_crash' goes
 *	  to zero (this is a countdown counter do allow for setting
 *	  it with a time-delay).  Normally, it is just set to
 *	  1 with a kernel debugger, and a system panic happens in the
 *	  next 1 minute.  For some deadlocks, it may not be possible
 *	  to run a kernel debugger after the deadlock has happened,
 *	  so this flag can be set to a large value (i.e. 60) and
 *	  then the test that deadlocks is run, and then you wait 60 minutes
 *	  for it to panic with the deadlock case.
 *	- Optionally does a scan for internal consistency of the mnode
 *	  database i.e. the call to MDB_CHKPOINT(1)
 *	- Displays the heap on the console if KMEM tracking is on and 
 *	  the flag is set to print the heap (used for debug of memory leaks)
 *	- Call mfs_procpurge() to garbage collect mfs proc structs
 *	  for any dead processes.
 *	- Calls mfs_vwdircleanhm() to timeout any history mode 
 *	  view-tags we created dynamically after a suitable time-period.
 */

void
mvfs_periodic_maintenance(CALL_DATA_T *cd)
{
    VNODE_T *rootvp;
    mvfs_common_data_t *zmcdp;
    int z;
    int vr_idx;

    ASSERT(MDKI_ATOMIC_READ_UINT32(&mvfs_state) == MVFS_SYNC_IN_PROGRESS);

    /*
     * Exit quickly on panic to avoid hangs that prevent
     * dumps from happening.
     */

    if (MDKI_PANICSTR()) return;

    /*
     * Check for debug crash requested.
     * The only way to set this flag is for 'root'
     * to patch the kernel with a debugger.
     * mdb_crash can be used for "delayed" crashes
     * by setting it to a value greater than 1;
     * it will be decremented on each sync and
     * panic on the transition to 0.
     */

    if ((mdb_crash != 0) && (--mdb_crash == 0)) {
        MDKI_PANIC("MVFS debug crash");
    }

    MDB_CHKPOINT(1);    /* debug scan in background */

    MFS_PRKMEM(); /* Display heap if flag set && using mvfs kmem */

    mvfs_procpurge(MVFS_PROCPURGE_NOSLEEP); /* Clean up dead processes */
    MVFS_FLUSH_CREDLIST(FALSE);

    /* Clean up stale HM views in all the viewroots (automatic and dynamic). */
    for (vr_idx = 0; vr_idx < MVFS_VIEWROOT_SZ; vr_idx++) {
        /* Don't use MVFS_GETVIEWROOT here. MVFS_GETVIEWROOT attempts to lock
         * the mount_lock which could cause a deadlock if the last viewroot
         * is being unmounted.  It is safe here to directly access this field
         * because the only place it is destroyed in is mvfs_vunmount_subr,
         * which first sets mfs_viewroot_vp to NULL and then waits until
         * mvfs_state != MVFS_SYNC_IN_PROGRESS. So, we'll either pick up NULL
         * here and skip syncing or get a vnode which will remain valid until
         * mvfs_state is switched back to MVFS_FULLY_INITIALIZED (e.g. when this
         * function returns to mvfs_vsync).
         */
        rootvp = MDKI_VIEWROOT_GET_DATAP(vr_idx)->mfs_viewroot_vp;

        if (rootvp != NULL) {
            VN_HOLD(rootvp); /* Just to be safe. */
            mfs_viewdircleanhm(rootvp, cd); /* FIXME: needs nowait option! */
            ATRIA_VN_RELE(rootvp, cd);
        }
    }

    return;
}

/*
 * MFS_SYNC - vnode version of sync call.
 * This is called every 30 secs or minute.  Use it for background cleanup
 * work.  Note that this is called with vfsp = NULL to indicate all
 * the vfs's should be cleaned up.
 *
 */

int
mvfs_vsync(
    VFS_T *vfsp,
    mvfs_vsync_flags_t flags,
    CALL_DATA_T *cd
)
{
    int error = 0;
    mvfs_viewroot_data_t *vrdp = NULL;

    /* Exit quickly on panic to avoid hangs that prevent
     * dumps from happening.
     */
    if (MDKI_PANICSTR()) {
        return(0);
    }

    MDKI_ATOMIC_INCR_UINT32(&mvfs_us_ops_in_progress);

retry_sw:
    switch (MDKI_ATOMIC_READ_UINT32(&mvfs_state)) {
      case MVFS_FULLY_INITIALIZED:
        if (!MDKI_ATOMIC_CAS_UINT32(&mvfs_state,
                                    MVFS_FULLY_INITIALIZED,
                                    MVFS_SYNC_IN_PROGRESS))
        {
            goto retry_sw;
        }
        break;
      case MVFS_SYNC_IN_PROGRESS:
        MDKI_USECDELAY(50000);
        goto retry_sw;
      default:
        MDKI_ATOMIC_DECR_UINT32(&mvfs_us_ops_in_progress);
        return(0);
    }

    /* Do periodic maintenance functions off the sync
     * operations on Unix.  That way we don't have to
     * worry about creating system threads or deferred
     * procedure calls in the Unix kernel.
     * (Callouts via timeout() usually happen at an
     * elevated interrupt priority level at which the
     * periodic maintenance operations can not be done,
     * so we can't just use the timeout() mechanism in the
     * kernel).
     */
    if (vfsp == NULL) {
        mvfs_periodic_maintenance(cd);
    } else {
        vrdp = MDKI_VIEWROOT_GET_DATAP(MVFS_VRTYPE_VFS(vfsp));
        if ((vrdp != NULL) &&
            (vrdp->mfs_viewroot_vfsp == NULL ||
             vfsp == vrdp->mfs_viewroot_vfsp))
        {
            /* Only do this once per sync call.  Some systems call us only
             * once with a null vfsp.  Other systems call us for each vfs,
             * so only do this once if the viewroot is mounted (and once in
             * the odd case that the viewroot isn't mounted).
             */
            mvfs_periodic_maintenance(cd);
        }
    }

    /* Log first, usually have to debug a hang here ... */

    MDB_VFSLOG((MFS_VSYNC,
                "vfsp=%"KS_FMT_PTR_T" time= %"MVFS_FMT_CTIME_X"\n",
                vfsp, MDKI_CTIME()));
    BUMPSTAT(mfs_vfsopcnt[MFS_VSYNC]);

    /* Sync any "dirty" mnodes (vob mnodes with cached pages)
     */
    if (mfs_do_sync && MVFS_FLAGOFF(flags, MVFS_VSYNC_NO_MNODE_SYNC)) {
        mvfs_mnsyncmnodes(vfsp, cd);
    }

cleanup:
    if (!MDKI_ATOMIC_CAS_UINT32(&mvfs_state,
                                MVFS_SYNC_IN_PROGRESS,
                                MVFS_FULLY_INITIALIZED))
    {
        MDKI_PANIC("mvfs_vsync: mvfs_state not MVFS_SYNC_IN_PROGRESS "
                   "on exit.\n");
    }

    MDKI_ATOMIC_DECR_UINT32(&mvfs_us_ops_in_progress);

    return(0);
}

/*
 * MFS_GETMNODE - find a VOB mnode from a view vnode and internal fid
 *
 * This routine leaves the MNODE REFCOUNT INCREMENTED AND RETURNS
 * with the mnode still locked!.  Callers of
 * this routine are responsible for calling mfs_mnrele when done
 * with the mnode, or when the vnode refcount has been successfully
 * incremented.
 * The comments said that this could be called with a NULL cred (cd) 
 * pointer.  This may be true for platforms that have not converted
 * to using call_data structures.  Platforms making a conversion need
 * to make sure that they pass call_data, even through mfs_vget or
 * their equivalent.
 */

int
mfs_getmnode(
    VFS_T *vfsp,
    VNODE_T *vw,
    mfs_fid_t *fidp,
    mfs_mnode_t **mnpp,
    int *nnp,		/* Newnode flag ptr */
    CALL_DATA_T *cd
)
{
#ifndef MVFS_WRAP_ENTER_EXIT
    CRED_T *cred;
#endif
    int error = 0;
    mfs_mnode_t *mnp;

    /* Could we get a vnode's vp->v_vfsp which is a "local" file
     * system VFS type by accident, e.g. on systems which manufacture
     * LOOPCLAS vnodes with vp->v_vfsp pointing at the real vfsp, but
     * vp->v_ops pointing at MVFS?  ASSERT that such a case isn't
     * showing up in this routine.
     */
    ASSERT(MVFS_VFSISMVFS(vfsp, mfs_vfsopp));

    if (MVFS_ISVIEWROOTMNT(VFS_TO_MMI(vfsp))) {
        /* Never allow viewroot nodes back to NFS daemon (the only
           caller). */
        error = ESTALE;
        mnp = NULL;
    } else { 
        if (fidp->mf_dbid == MVFS_ROOTDBID) {
	    if (fidp->mf_gen != MVFS_ROOTGEN) {
		error = ESTALE;
		mnp = 0;
	    } else {
		mnp = mfs_mnget(MFS_VOBRTCLAS, vw, fidp, vfsp, nnp);
		if (!mnp) error = ESTALE;
	    }
	} else {
	    if (vw != NULL) {
	        mnp = mfs_mnget(MFS_VOBCLAS, vw, fidp, vfsp, nnp);
		if (!mnp) error = ESTALE;
	
	        /* New VOB node, must get info via RPC to fill it in */

	        if (!error && *nnp) {

		    /* Make partial view file handle */

		    MFS_FID_TO_PARTIAL_VFH(*fidp, vfsp, vw, mnp->mn_vob.vfh);

		    /* 
		     * Get attributes (and full vfh) from view 
		     * For non-atria access, the credentials may
		     * be NULL because the vget() op does not have
		     * credentials as an argument.  In addition, the process
		     * and/or thread credentials may not be set to the
		     * actual user, but be left at sys_cred.
		     * This can create problems if the view ever did an
		     * access check on the creds for a getattr!
                     * With the conversion to using call_data instead of
                     * creds, converted platforms will always pass a cd
                     * pointer.  
                     */

#ifndef MVFS_WRAP_ENTER_EXIT
                    cred = MVFS_CD2CRED(cd);
                    if (cred == NULL) {
                        cred = MVFS_DUP_DEFAULT_CREDS();
                        if (cred == NULL) {
                            error = ENOMEM;
                        } else {
                            error = mfs_clnt_getattr_mnp(mnp, vfsp, cred);
                            MDKI_CRFREE(cred);
                            cred = NULL;
                        }
                    } else {
                        error = mfs_clnt_getattr_mnp(mnp, vfsp, cred);
                    }
#else
                    error = mfs_clnt_getattr_mnp(mnp, vfsp, cd);
#endif
		    BUMPSTAT(mfs_acstat.ac_misses);
		    BUMPSTAT(mfs_acstat.ac_newmiss);

		    /* 
		     * If no error, and only have partial fid, promote vfh gen
                     * back up to fid. (VFH was filled in by clnt_getattr) 
		     * If there is an error, then mark this mnode "stale"
                     * and clear the fid so it won't be reused with bad
                     * data in it, and then free it up and return a NULL
                     * mnode ptr to the client.  Don't forget to unlock!
		     */

	 	    if (!error) {
			if (mnp->mn_hdr.fid.mf_gen == 0)
		            mnp->mn_hdr.fid.mf_gen = mnp->mn_vob.vfh.gen;
		    } else {
			mnp->mn_hdr.stale = 1;
			
			mnp->mn_hdr.fid.mf_gen = MFS_NULL_GEN;
			MUNLOCK(mnp);
			mfs_mnrele(mnp, cd);
			mnp = NULL;
		    }
		}	/* if (*nnp) ... */
	    } else {	/* vw != NULL... else ...  */
	        mnp = NULL;
		error = ESRCH;
	    }
	}		/* fidp == ROOTDBID... else ... */
    }			/* VIEWSVRMNT ... else ... */

    /* NOTE: returns with mnode locked, and mnode refcount incremented. */
    *mnpp = mnp;
    return(error);
}

int
mfs_getvnode(
    VFS_T *vfsp,
    VNODE_T *vw,
    mfs_fid_t *fidp,
    VNODE_T **vpp,
    CALL_DATA_T *cd
)
{
    int error = 0;
    mfs_mnode_t *mnp;
    VNODE_T *vp = NULL;
    int newnode;
    VTYPE_T vtype;

    /*
     * Get the mnode from the FID.  Note that this returns with
     * mnode locked and mnode refcount incremented.
     */

    /*
     * All internal code is calling this in a VOB context (vfsp is for
     * a VOB mount).  NFS server might call it with a viewroot, which
     * we don't support, but let getmnode bounce that.
     */
    error = mfs_getmnode(vfsp, vw, fidp, &mnp, &newnode, cd);
    if (!error) {
	ASSERT(mnp->mn_hdr.mcount > 0);
    }

    /*
     * Get the vnode.  
     * See mvfs_makevobnode in vnodeops for a discussion of mnode and
     * vnode refcounts and how they are used.  This routine
     * decrements mnode refcount (if appropriate) and unlocks the
     * mnode.
     */

    if (!error) {
        error = MVFS_VNGET(vfsp, NULL, mnp, &vp, cd);
    }

    *vpp = vp;
    return(error);
}

/*
 * MFS_UUID_TO_HASH32 - utility routine to hash a uuid
 * Really a copy of tbs_uuid_to_hash32, but for the kernel.
 */

u_long
mfs_uuid_to_hash32(uuid_p)
A_CONST tbs_uuid_t *uuid_p;
{
    ks_uint32_t val;
    /* note: only the high part of the low time varies greatly */
    /* -- and then only when time has past */
    val = (uuid_p->time_low >> 2) + (uuid_p->time_low >> 22) +
            ((uuid_p->time_mid + uuid_p->time_hi_and_version
            + uuid_p->clock_seq_low
            + (uuid_p->clock_seq_hi_and_reserved << 26)
            + (uuid_p->node[5] << 20) + (uuid_p->node[4] << 16)
            + (uuid_p->node[3] << 12) + (uuid_p->node[2] << 8)
            + (uuid_p->node[1] << 4) + uuid_p->node[0])
        * 536870909); /* large prime */
    return (long) (val & 0xFFFFFFFF);		/* mask to 32 bits */
}

/*
 * MFS_VGET - find a vnode from its external (NFS) fid
 * This is now a wrapper around mvfs_vget_cd which expects to receive
 * a call_data block.  This is to preserve compatability with unconverted
 * platforms where the interface is controlled by the base OS.
 */
int
mfs_vget(
    VFS_T *vfsp,
    VNODE_T **vpp,
    struct fid *xfidp
)
{
    return mvfs_vget_cd(vfsp, vpp, xfidp, NULL);
}

int
mvfs_vget_cd(
    VFS_T *vfsp,
    VNODE_T **vpp,
    struct fid *xfidp,
    CALL_DATA_T *cd
)
{
    int error = 0;
    mfs_xfid_t *mxfidp;
    mfs_fid_t fid;
    VNODE_T *vw;
    mfs_mnode_t *mnp;
    int exportid;
    mvfs_viewroot_data_t *vrdp;

    /* If not a vob mount return EINVAL */
    if (MVFS_ISVIEWROOTMNT(VFS_TO_MMI(vfsp))) {
        return(EINVAL);
    }
    /* This macro does some redundant checking for a viewroot, which we just
    ** checked above, but it's pretty cheap, and makes things clearer.
    */
    vrdp = MDKI_VIEWROOT_GET_DATAP(MVFS_VRTYPE_VFS(vfsp));

    /* Cast to mfs fid structure */

    mxfidp = (mfs_xfid_t *)xfidp;

    /* Get view if a viewroot vfs exists and valid view id */

    if (vrdp->mfs_viewroot_vfsp && mxfidp->mfx_vid != MFS_NULLVID) {
	mnp = VTOM(vrdp->mfs_viewroot_vp);
	MLOCK(mnp);
	if (mxfidp->mfx_vid < (unsigned) mnp->mn_ramdir.export_hwm &&
	    mnp->mn_ramdir.export_ents[mxfidp->mfx_vid] != -1) {
	    exportid = mnp->mn_ramdir.export_ents[mxfidp->mfx_vid];
	    vw = mnp->mn_ramdir.ents[exportid].vp;
            /* Hold vw in case someone else releases it from
               ramdir/export table */
            VN_HOLD(vw);
	} else
	    vw = NULL;
	MUNLOCK(mnp);
    } else {
	vw = NULL;
    }

    /* Build normal fid from rest of export fid */
  
    fid.mf_dbid = mxfidp->mfx_dbid;
    fid.mf_gen  = mxfidp->mfx_gen;

    /* Restore real gen by subtracting the view uuid hash
     * we added in on the fetch of the fid.  If the view-id
     * has accidently hooked us back up to the wrong view,
     * then this will create a bogus generation number and
     * we will get ESTALE.
     */

    if (vw) {
	fid.mf_gen -= 
	    mfs_uuid_to_hash32(&(VTOM(vw)->mn_view.svr.uuid));
    }
    /*
     * also subtract out VOB uuid hash (needed to prevent same view/different
     * VOB confusion at the unbound VOB root)
     */
    fid.mf_gen -= mfs_uuid_to_hash32(&(VFS_TO_MMI(vfsp)->mmi_svr.uuid));

    error = mfs_getvnode(vfsp, vw, &fid, vpp, cd);
    if (vw != NULL) {
        ATRIA_VN_RELE(vw, cd);
    }
    /* 
     * I would like to fetch the cleartext here for non-atria
     * access, but unfortuneately, this op has no credentials
     * associated with it for the lookup.
     * Thus the getcleartext has to be deferred to read/write/getattr
     * ops (as required depending on system implementation).
     *
     */

    MDB_VFSLOG((MFS_VGET, "vfsp=%"KS_FMT_PTR_T" *vpp=%"KS_FMT_PTR_T", fid=%x/%x/%x/%x err=%d\n", vfsp,
		*vpp, ((int *)mxfidp)[0], ((int *)mxfidp)[1],
		((int *)mxfidp)[2], ((int *)mxfidp)[3], error));
    BUMPSTAT(mfs_vfsopcnt[MFS_VGET]);
    MDB_CHKPOINT(0);
    return(error);
}

int
mfs_nosys(void)
{
    mvfs_log(MFS_LOG_WARN, "unimplemented vfs or vnode op\n");
    return(ENOSYS);
}

int
mfs_swapvp(void)
{
    mvfs_log(MFS_LOG_WARN, "unimplemented swapvp vfs op\n");
    return(ENOSYS);
}

/* Allocate a minor device number */

int
mvfs_vfsgetnum(map, size)
MVFS_MINMAP_T *map;
int size;
{
    u_char *p;
    u_char *ep;
    unsigned val;
    int i;
    SPL_T s;

    SPLOCK(map->lock, s);
    for (p = &map->vec[0], ep = &map->vec[size]; p < ep; p++) {
	if (*p != (u_char)(~0)) {
	    val = *p;
	    for (i = 0; i < NBBY; i++) {
		if ((val & (1 << i)) == 0) {
		    *p |= (1 << i);
		    SPUNLOCK(map->lock, s);
		    return(int)(((p - (&map->vec[0])) * NBBY) + i);
		}
	    }
	    MDKI_PANIC("mvfs: mfsvfsgetnum: bad map");
	}
    }
    SPUNLOCK(map->lock, s);
    return(-1);
}

/*
 * Allocate a particular minor device number pair, returning an error
 * if not possible.
 */
int
mvfs_get_minorpair(map, size, index)
MVFS_MINMAP_T *map;
int size;
int index;
{
    u_char *p;
    unsigned val;
    int i, bitoffset;
    SPL_T s;

    if (index > size * NBBY)
	return 1;			/* out of range */
    if ((index & 1) != 0)
	return 1;			/* not an even starting index--
					 * this means we can't check both bits
					 * in same byte, yech!
					 */
    ASSERT((NBBY / 2)*2 == NBBY);	/* must be even # bits in a byte */
    SPLOCK(map->lock, s);
    p = &map->vec[index/NBBY];
    bitoffset = index % NBBY;

    if ((*p & (3 << bitoffset)) == 0) {
	/* available. */
	*p |= (3 << bitoffset);
	SPUNLOCK(map->lock, s);
	return 0;
    }
    SPUNLOCK(map->lock, s);
    return 1;				/* no room, sorry */
}

int
mvfs_vfsputnum(map, size, n)
    MVFS_MINMAP_T *map;
    int size;
    int n;
{
    int ix, off;
    SPL_T s;

    if (n >= (size * NBBY))
	MDKI_PANIC("mvfs: mfs_vfsputnum: out of range");
    ix = n / NBBY;
    off = n - (ix * NBBY);
    SPLOCK(map->lock, s);
    if ((map->vec[ix] & (1 << off)) == 0)
	MDKI_PANIC("mvfs: mfs_vfsputnum: not allocated");
    map->vec[ix] &= ~(1 << off);
    SPUNLOCK(map->lock, s);
    return(0);
}

/*
 * MFS_GETVIEWROOT - get a held vnode ptr to the viewroot
 */

VNODE_T *
mvfs_getviewroot(mvfs_viewroot_type_t vrtype)
{
    VNODE_T *dvp;
    mvfs_vfs_data_t *mvdp = MDKI_VFS_GET_DATAP(NULL);
    mvfs_viewroot_data_t *vrdp = MDKI_VIEWROOT_GET_DATAP(vrtype);

    MVFS_LOCK(&(mvdp->mount_lock));  /* Freeze the mount table */
    if ((dvp = vrdp->mfs_viewroot_vp) != NULL) {
        VN_HOLD(dvp);           /* Hold view root */
    }
    MVFS_UNLOCK(&(mvdp->mount_lock));
    return(dvp);
}

/*
 * Convert mmi_minor in struct mfs_mntinfo to the major/minor number actually
 * used by OS. mmi_minor may be larger than MVFS_MINORMAX, so we need to split
 * it into a major/minor combination.
 */
STATIC int
mvfs_find_devnum(
    int mmi_minor, 
    dev_t *majorp, 
    int *minorp
)
{
    int major_index;
    major_index = mmi_minor / MVFS_MINORMAX;
    if (mvfs_majortbl[major_index] == -1) {
        if ((mvfs_majortbl[major_index] = MDKI_GET_MAJOR_DEV()) == -1)
            return ENODEV;
    }
    *majorp = mvfs_majortbl[major_index];
    *minorp = mmi_minor % MVFS_MINORMAX;
    return 0;
}

/*
 * Search in vob mount list, given a comparison callback.
 * Returns the result of eval_func or NULL if nothing found.
 * NB: The eval_func callback is called with mount_lock locked,
 * and if the eval_func returns a result, that result will be
 * returned to the caller after the mount_lock is unlocked.
 * Therefore, if the result is, e.g. a vfs pointer, then the
 * structure should be held before the eval_func returns it, if
 * necessary.
 *
 * XXXODFS: I think this function should only ever be called to look for "real"
 * vfs's...there's no view passed in, so we wouldn't know what per-view data to
 * use for an automatic view.
 */
void *
mvfs_find_mount(
    void *(* eval_func)(VFS_T *vfsp, void *data),
    void *data
)
{
    int i;
    void *result  = NULL;
    mvfs_vfs_data_t *mvdp = MDKI_VFS_GET_DATAP(NULL);

    MVFS_LOCK(&(mvdp->mount_lock));

    /*
     * searching backwards because nca_exported are
     * placed into the end of the array
     */
    for (i = mvdp->vob_mount_hwm; i >= 0; i--) {
        if (mvdp->vob_mounts[i] != NULL) {
            result = eval_func(mvdp->vob_mounts[i], data);
            if (result != NULL)
                break;
        }
    }
    MVFS_UNLOCK(&(mvdp->mount_lock));
    return result;
}
static const char vnode_verid_mvfs_vfsops_c[] = "$Id:  ab97371e.5f1211e6.9d34.00:01:84:b0:ce:b2 $";
